# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from pyiron_contrib.protocol.utils.misc import getLogger
from abc import ABCMeta
import sys

"""
Utility function to inject subclasses of pyiron_contrib.protocol.generic.Protocol as JobTypes into pyiron
"""

__author__ = "Dominik Gehringer, Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "December 10, 2019"


# we have to use ABC as parent metatype otherwise we will run into a metaclass conflict however for Protocol that is
#exactly what we want since one is not allowed to instantiate a protocol
class PyironJobTypeRegistryMetaType(ABCMeta):
    """
    Metatype which keeps track of all its subclasses
    """

    __registry = {}
    __wrapped = {}

    @staticmethod
    def inject_dynamic_types():
        """
        Sets all the dynimically created classes as attributes to the current module object, however with the original
        name. Updates pyiron JobTypeChoice's JOB_CLASS_DICT
        """

        # fetch the current module name
        module_name = PyironJobTypeRegistryMetaType.__module__

        if module_name not in sys.modules:
            # throw a info message if the module is no loaded yes -> could possibly happen but is very unlikely
            getLogger('PyironJobTypeRegistryMetaType').info('Skipping not module "%s" not loaded yet. Please call PyironJobTypeRegistryMetaType.inject_dynamic_types manually')
        else:
            # get the module object
            module = sys.modules[module_name]
            updates = {}
            # register all dynamically created subclasses of (cls, GenericJob) as attributes
            for (origin_module, wrapped_name), tp in PyironJobTypeRegistryMetaType.__wrapped.items():
                if not hasattr(module, wrapped_name):
                    setattr(module, wrapped_name, tp)
                    # information needed for pyiron.base.job.jobtype.JOB_CLASS_DICT
                    updates[wrapped_name] = module_name

            # inject it into pyiron
            # adding it to JOB_CLASS_DICT also enables it for autocomplete
            from pyiron_base import JOB_CLASS_DICT
            # make sure that it affects all instances of pyiron.base.job.jobtype.JobTypeChoice
            JOB_CLASS_DICT.update(updates)

    def __init__(cls, name, bases, nmspc):
        """
        type initialize for type of this metaclass
        Args:
            name: (str) class name
            bases: (tuple of type) the base classes
            nmspc: (dict) the attribute of the tpye
        """

        super(PyironJobTypeRegistryMetaType, cls).__init__(name, bases, nmspc)
        # that's a little bit ugly but we want to exclude this types and do not want to wrap them
        if name in ('PyironJobTypeRegistry', 'Protocol', 'PyironJobTypeRegistryMetaType'):
            return
        # convenience function to obtain the full class qualifier
        fullname = lambda : '%s.%s' %( cls.__module__, cls.__name__)

        class_name = cls.__name__
        class_module = cls.__module__
        key = (class_module, class_name)
        # construct the new class name
        new_class_name = class_name
        # we place a `__artificial__` attribute in the type information to indicate that it is a dynamically created type
        # when calling `type(new_class_name, new_bases, new_spec)` we construct a subclass of cls, thus it is a
        # recsursive call to this method, we have to avoid that
        # otherwise we wrap the wrapper again and agin
        if '__artificial__' not in nmspc:
            from pyiron_base import GenericJob
            new_bases = (cls, GenericJob)
            # edit the metainformation
            new_spec = nmspc.copy()
            new_spec['__qualname__'] = new_class_name
            new_spec['__artificial__'] = True
            new_spec['__module__'] = PyironJobTypeRegistryMetaType.__module__
            # this bad guy caused some troubles
            if '__classcell__' in new_spec:
                del new_spec['__classcell__']
            try:
                new_type = type(new_class_name, new_bases, new_spec)
            except TypeError as e:
                # if creating the type goes wrong we are really screwed up
                getLogger(fullname()).exception(fullname(), exc_info=e)
                raise
            # register everything nicely
            PyironJobTypeRegistryMetaType.__registry[key] = cls
            PyironJobTypeRegistryMetaType.__wrapped[key] = new_type
            PyironJobTypeRegistryMetaType.wrapped = PyironJobTypeRegistryMetaType.__wrapped

            # try to inject it -> in case of dynamically created subclasses
            PyironJobTypeRegistryMetaType.inject_dynamic_types()


class PyironJobTypeRegistry(metaclass=PyironJobTypeRegistryMetaType):
    """
    Convenience class as it is ABC in the abc module
    """
    pass



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

from pyiron_contrib.protocol.generic import CompoundVertex, Protocol
from pyiron_contrib.protocol.primitive.one_state import Counter, ExternalHamiltonian, GradientDescent, \
    NEBForces, InterpolatePositions
from pyiron_contrib.protocol.primitive.two_state import IsGEq
from pyiron_contrib.protocol.list import SerialList, ParallelList, AutoList
from pyiron_contrib.protocol.utils import Pointer
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

"""
Protocol for nudged elastic band minimization.
"""

__author__ = "Liam Huber, Raynol Dsouza, Jan Janssen"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "18 July, 2019"


class NEB(CompoundVertex):
    """
    Relaxes a system according to the nudged elastic band method (Jonsson et al).

    Input dictionary:
        ref_job_full_path (str): Path to the pyiron job to use for evaluating forces and energies.
        structure_initial (Atoms): The starting structure for the elastic band.
        structure_final (Atoms): The final structure for the elastic band. Warning: must have the same numper and
            species of atoms, cell, and periodic boundary conditions as the initial structure.
        n_images (int): How many images create by interpolating between the two structures.
        n_steps (int): How many steps to run for. (Default is 100.)
        f_tol (float): Ionic force convergence (largest atomic force). (Default is 1e-4 eV/angstrom.)
        spring_constant (float): Spring force between atoms in adjacent images. (Default is 1.0 eV/angstrom^2.)
        tangent_style ('plain'/'improved'/'upwinding'): How to calculate the image tangent in 3N-dimensional space.
            (Default is 'upwinding'.)
        use_climbing_image (bool): Whether to replace the force with one that climbs along the tangent direction for
            the job with the highest energy. (Default is True.)
        smoothing (float): Strength of the smoothing spring when consecutive images form an angle. (Default is None, do
            not apply such a force.)
        gamma0 (float): Initial step size as a multiple of the force. (Default is 0.1.)
        fix_com (bool): Whether the center of mass motion should be subtracted off of the position update. (Default is
            True)
        use_adagrad (bool): Whether to have the step size decay according to adagrad. (Default is False)

    Output dictionary:
        energy_pot (list[float]): Total potential energy of the system in eV.
        positions_list (list[numpy.ndarray]): Atomic positions in angstroms for each image.
        forces_list (list[numpy.ndarray]): Atomic forces in eV/angstrom for each image, including spring forces.
    """

    DefaultWhitelist = {
        'calc_static': {
            'output': {
                'energy_pot': 1,
            },
        },
    }

    def __init__(self, **kwargs):
        super(NEB, self).__init__(**kwargs)

        # Protocol defaults
        id_ = self.input.default
        id_.n_steps = 100
        id_.f_tol = 1e-4

        id_.spring_constant = 1.
        id_.tangent_style = 'upwinding'
        id_.use_climbing_image = True
        id_.smoothing = None

        id_.gamma0 = 0.1
        id_.fix_com = True
        id_.use_adagrad = False

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.interpolate_images = InterpolatePositions()
        g.check_steps = IsGEq()
        g.calc_static = AutoList(ExternalHamiltonian)
        g.neb_forces = NEBForces()
        g.gradient_descent = SerialList(GradientDescent)
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.interpolate_images,
            g.check_steps, 'false',
            g.calc_static,
            g.neb_forces,
            g.gradient_descent,
            g.clock,
            g.check_steps
        )
        g.starting_vertex = self.graph.interpolate_images
        g.restarting_vertex = self.graph.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        g.interpolate_images.input.structure_initial = ip.structure_initial
        g.interpolate_images.input.structure_final = ip.structure_final
        g.interpolate_images.input.n_images = ip.n_images

        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        g.calc_static.input.n_children = ip.n_images
        g.calc_static.direct.ref_job_full_path = ip.ref_job_full_path
        g.calc_static.direct.structure = ip.structure_initial
        g.calc_static.broadcast.default.positions = gp.interpolate_images.output.interpolated_positions[-1]
        g.calc_static.broadcast.positions = gp.gradient_descent.output.positions[-1]

        g.neb_forces.input.default.positions_list = gp.interpolate_images.output.interpolated_positions[-1]
        g.neb_forces.input.positions_list = gp.gradient_descent.output.positions[-1]
        g.neb_forces.input.energies = gp.calc_static.output.energy_pot[-1]
        g.neb_forces.input.forces_list = gp.calc_static.output.forces[-1]
        g.neb_forces.input.cell = ip.structure_initial.cell
        g.neb_forces.input.pbc = ip.structure_initial.pbc
        g.neb_forces.input.spring_constant = ip.spring_constant
        g.neb_forces.input.tangent_style = ip.tangent_style
        g.neb_forces.input.use_climbing_image = ip.use_climbing_image
        g.neb_forces.input.smoothing = ip.smoothing

        g.gradient_descent.input.n_children = ip.n_images
        g.gradient_descent.broadcast.default.positions = gp.interpolate_images.output.interpolated_positions[-1]
        g.gradient_descent.broadcast.positions = gp.gradient_descent.output.positions[-1]
        g.gradient_descent.broadcast.forces = gp.neb_forces.output.forces_list[-1]
        g.gradient_descent.direct.masses = ip.structure_initial.get_masses
        g.gradient_descent.direct.gamma0 = ip.gamma0
        g.gradient_descent.direct.fix_com = ip.fix_com
        g.gradient_descent.direct.use_adagrad = ip.use_adagrad

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'energy_pot': ~gp.calc_static.output.energy_pot[-1],
            'positions_list': ~gp.gradient_descent.output.positions[-1],
            'forces_list': ~gp.neb_forces.output.forces_list[-1]
        }

    def _get_energies(self, frame=None):
        if frame is None:
            return self.graph.calc_static.output.energy_pot[-1]
        else:
            return self.graph.calc_static.archive.output.energy_pot[frame]

    def plot_elastic_band(self, ax=None, frame=None, plot_kwargs=None):
        if ax is None:
            _, ax = plt.subplots()
        if plot_kwargs is None:
            plot_kwargs = {}
        if 'marker' not in plot_kwargs.keys():
            plot_kwargs = {'marker': 'o'}
        energies = np.array(self._get_energies(frame=frame))
        ax.plot(energies - energies[0], **plot_kwargs)
        ax.set_ylabel("Energy")
        ax.set_xlabel("Image")
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        return ax

    def get_forward_barrier(self, frame=None):
        """
        Get the energy barrier from the 0th image to the highest energy (saddle state).
        """
        energies = self._get_energies(frame=frame)
        return np.amax(energies) - energies[0]

    def get_reverse_barrier(self, frame=None):
        """
        Get the energy barrier from the final image to the highest energy (saddle state).
        """
        energies = self._get_energies(frame=frame)
        return np.amax(energies) - energies[-1]

    def get_barrier(self, frame=None):
        return self.get_forward_barrier(frame=frame)


class NEBSerial(NEB):
    """
    The same as the NEB protocol, but ensures that the force/energy calls are treated in series.
    """

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.interpolate_images = InterpolatePositions()
        g.check_steps = IsGEq()
        g.calc_static = SerialList(ExternalHamiltonian)  # Enforce serial
        g.neb_forces = NEBForces()
        g.gradient_descent = SerialList(GradientDescent)
        g.clock = Counter()


class NEBParallel(NEB):
    """
    The same as the NEB protocol, but ensures that the force/energy calls are treated in parallel.
    """

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.interpolate_images = InterpolatePositions()
        g.check_steps = IsGEq()
        g.calc_static = ParallelList(ExternalHamiltonian)  # Enforce parallel
        g.neb_forces = NEBForces()
        g.gradient_descent = SerialList(GradientDescent)
        g.clock = Counter()


class ProtocolNEB(Protocol, NEB):
    pass



__version__ = "0.1"
__all__ = []

from pyiron import Project
from pyiron_base import JOB_CLASS_DICT

# Make classes available for new pyiron version
JOB_CLASS_DICT['ProtocolMinimize'] = 'pyiron_contrib.protocol.compound.minimize'
JOB_CLASS_DICT['ProtocolMD'] = 'pyiron_contrib.protocol.compound.molecular_dynamics'
JOB_CLASS_DICT['ProtocolNEB'] = 'pyiron_contrib.protocol.compound.nudged_elastic_band'
JOB_CLASS_DICT['ProtocolQMMM'] = 'pyiron_contrib.protocol.compound.qmmm'
JOB_CLASS_DICT['ProtocolHarmonicTILD'] = 'pyiron_contrib.protocol.compound.tild'
JOB_CLASS_DICT['ProtocolHarmonicTILDParallel'] = 'pyiron_contrib.protocol.compound.tild'
JOB_CLASS_DICT['ImageJob'] = 'pyiron_contrib.image.job'

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from enum import Enum
from pyiron_contrib.protocol.utils.misc import  LoggerMixin, requires_arguments
from types import MethodType, FunctionType

"""
Python implementation of pointers. Pointer can be resolved using ~ operator
"""

__author__ = "Dominik Gehringer, Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "December 10, 2019"

class CrumbType(Enum):
    """
    Enum class. Provides the types which Crumbs in an IODictionary are allowed to have
    """

    Root = 0
    Attribute = 1
    Item = 2


class Crumb(LoggerMixin):
    """
    Represents a piece in the path of the IODictionary. The Crumbs are used to resolve a recipe path correctly
    """

    def __init__(self, crumb_type, name):
        """
        Initializer from crumb
        Args:
            crumb_type (CrumbType): the crumb type of the object
            name (str, object): An object if crumb type is CrumbType.Root otherwise a string
        """
        self._crumb_type = crumb_type
        self._name = name

    @property
    def name(self):
        return self._name

    @property
    def crumb_type(self):
        return self._crumb_type

    @property
    def object(self):
        if self.crumb_type != CrumbType.Root:
            raise ValueError('Only root crumbs store an object')
        return self._name

    @staticmethod
    def attribute(name):
        """
        Convenience method to produce an attribute crumb

        Args:
            name (str): the name of the attribute

        Returns:
            Crumb: the attribute crumbs

        """
        return Crumb(CrumbType.Attribute, name)

    @staticmethod
    def item(name):
        """
        Convenience method to produce an item crumb

        Args:
            name (str): the name of the item

        Returns:
            Crumb: the item crumb
        """
        return Crumb(CrumbType.Item, name)

    @staticmethod
    def root(obj):
        """
        Convenience method to produce a root crumb

        Args:
            obj: The root object of the path in the IODictionary

        Returns:
            Crumb: A root crumb, meant to be the first item in a IODictionary path
        """
        return Crumb(CrumbType.Root, obj)

    def __repr__(self):
        return '<{}({}, {})'.format(self.__class__.__name__,
                                    self.crumb_type.name,
                                    self.name if isinstance(self.name, str) else self.name.__class__.__name__)

    def __hash__(self):
        crumb_hash = hash(self._crumb_type)
        if self._crumb_type == CrumbType.Root:
            crumb_hash += hash(hex(id(self.object)))
        else:
            try:
                crumb_hash += hash(self._name)
            except Exception as e:
                self.logger.exception('Failed to hash "{}" object'.format(type(self._name).__name__), exc_info=e)
        return crumb_hash

    def __eq__(self, other):
        """
        Compares the crumb with an an object "other". Will return False if "other" is not an instance of Crumb
        Args:
            other: (object) the object to compare

        Returns: (bool) wether "other" is equal to self

        """
        if not isinstance(other, Crumb):
            return False
        if self.crumb_type == other.crumb_type:
            if self.crumb_type == CrumbType.Root:
                return hex(id(self.object)) == hex(id(other.object))
            else:
                return self.name == other.name
        else:
            return False


class Path(list, LoggerMixin):
    """
    A object representing a path to an objects attribute. It is a list of "Crumbs"
    The first object of always a Crumb of type CrumbType.Root followed by an arbitrary sequence of CrumbType.Item
    or CrumbType.Attribute
    """

    def append(self, item):
        if not isinstance(item, Crumb):
            raise TypeError('A path can only consist of crumbs')
        else:
            super(Path, self).append(item)

    def extend(self, collection):
        if not all([isinstance(item, Crumb) for item in collection]):
            raise TypeError('A path can only consist of crumbs')
        else:
            super(Path, self).extend(collection)

    def index(self, item, **kwargs):
        if not isinstance(item, Crumb):
            raise TypeError('A path can only consist of crumbs')
        else:
            return super(Path, self).index(item, **kwargs)

    def count(self, item):
        if not isinstance(item, Crumb):
            raise TypeError('A path can only consist of crumbs')
        else:
            return super(Path, self).count(item)

    @classmethod
    def join(cls, *p):
        return Path(p)


class Pointer(LoggerMixin):
    """
    A class pointing to an object. Can be resolved with ~p. It can be dangling at definition
    """

    def __init__(self, root):
        if root is not None:
            if not isinstance(root, Path):
                if not isinstance(root, Crumb):
                    path = [Crumb.root(root)]
                else:
                    path = [root]
            else:
                path = root.copy()
        else:
            raise ValueError('Root object can never be "None"')
        self.__path = path

    def __getattr__(self, item):
        return Pointer(Path.join(*self.__path, Crumb.attribute(item)))

    def __getitem__(self, item):
        return Pointer(Path.join(*self.__path, Crumb.item(item)))

    @property
    def path(self):
        return self.__path

    def _resolve_path(self):
        """
        This method resolves the object hiding behind a path (A list of Crumbs)

        Args:
            path (list<Crumb>): A list of path crumbs used to resolve the data object
            remaining (int): How many crumbs should be resolved

        Returns:
            The underlying data object, hiding behind path parameter

        """
        # Make a copy to ensure, because by passing by reference
        path = self.path.copy()

        # Have a look at the path and check that it starts with a root crumb
        root = path.pop(0)
        if root.crumb_type != CrumbType.Root:
            raise ValueError('Got invalid path. A valid path starts with a root object')
        # First element is always an object
        result = root.object
        while len(path) > 0:
            # Take one step in the path, pop the next crumb from the list
            crumb = path.pop(0)
            crumb_type = crumb.crumb_type
            crumb_name = crumb.name
            # If the result is a pointer itself we have to resolve it first
            if isinstance(result, Pointer):
                self.logger.info('Resolved pointer in a pointer path')
                result = ~result
            if isinstance(crumb_name, Pointer):
                self.logger.info('Resolved pointer in a pointer path')
                crumb_name = ~crumb_name
            # Resolve it with the correct method - dig deeper
            if crumb_type == CrumbType.Attribute:
                try:
                    result = getattr(result, crumb_name)
                except AttributeError as e:
                    self.logger.exception('Cannot fetch value "{}"'.format(crumb_name), exc_info=e)
                    raise e
            elif crumb_type == CrumbType.Item:
                try:
                    result = result.__getitem__(crumb_name)
                except (TypeError, KeyError) as e:
                    raise e

            # Get out of resolve mode
        return result

    def _resolve_function(self, function):
        """
        Convenience function to make IODictionary.resolve more readable. If the value is a function it calls the
        resolved functions if the do not require arguments

        Args:
            key (str): the key the value belongs to, just for logging purposes
            value (object/function): the object to resolve

        Returns:
            (object): The return value of the functions, if no functions were passed "value" is returned
        """
        result = function
        if isinstance(function, (MethodType, FunctionType)):
            # Make sure that the function has not parameters or all parameters start with
            if not requires_arguments(function):
                try:
                    # Get the return value
                    result = function()
                except Exception as e:
                    self.logger.exception('Failed to execute callable to resolve values for '
                                          'path {}'.format(self), exc_info=e)
                else:
                    self.logger.debug('Successfully resolved callable for path {}'.format(self))
            else:
                self.logger.warning('Found function, but it takes arguments! I \'ll not resolve it.')
        return result

    def __invert__(self):
        return self.resolve()

    def resolve(self):
        return self._resolve_function(self._resolve_path())



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import re
import numpy as np
from collections import OrderedDict
from pyiron_contrib.protocol.utils.pointer import Pointer
from pyiron_contrib.protocol.utils.misc import  LoggerMixin, fullname
from pyiron.atomistics.structure.atoms import Atoms

"""
Classes to setup input and output dataflows for protocols
"""

__author__ = "Dominik Gehringer, Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "December 10, 2019"

# define a regex to find integer values
integer_regex = re.compile(r'[-+]?([1-9]\d*|0)')


TIMELINE_DICT_KEY_FORMAT = 't_{time}'
GENERIC_LIST_INDEX_FORMAT = 'i_{index}'


class IODictionary(dict, LoggerMixin):
    """
    A dictionary class representing the parameters of a Command class. The dictionary holds a path which is recipe
    that can be resolved at runtime to obtain underlying values. A dictionary instance can hold multiple instances of
    IODictionary as value items which can be resolved into the real values when desired.
    """

    # those members are not accessible
    _protected_members = [
        'resolve',
        'to_hdf',
        'from_hdf'
    ]

    def __init__(self, **kwargs):
        super(IODictionary, self).__init__(**kwargs)

    def __getattr__(self, item):
        if item in IODictionary._protected_members:
            return object.__getattribute__(self, item)
        return self.__getitem__(item)

    def __getitem__(self, item):
        if item in self.keys():
            value = super(IODictionary, self).__getitem__(item)
            if isinstance(value, Pointer):
                return ~value
            elif isinstance(value, list):  # TODO: Allow other containers than list
                cls = type(value)
                return cls([element if not isinstance(element, Pointer) else ~element for element in value])
            else:
                return value
        return super(IODictionary, self).__getitem__(item)

    def __setattr__(self, key, value):
        super(IODictionary, self).__setitem__(key, value)

    def resolve(self):
        """
        Even though printing the dictionary, or asking for a particular item resolves paths fine, somehow using the **
        unpacking syntax fails to resolve pointers. This is to cover that issue since I couldn't find anything on
        google how to modify the ** behaviour.
        """
        resolved = {}
        for key in self.keys():
            resolved[key] = self.__getitem__(key)
        return resolved

    @staticmethod
    def _try_save_key(k, v, hdf, exclude=(dict, tuple, list)):
        """
        Tries to save a simple value

        Args:
            k: (str) key name of the HDF entry
            v: (obj) value
            hdf: the hdf server

        Returns: (bool) wether saving the key was successful or not

        """
        # try to call to_hdf first
        if hasattr(v, 'to_hdf'):
            v.to_hdf(hdf, group_name=k)
            result = True
        else:
            if exclude is not None:
                if isinstance(v, exclude):
                    return False
            try:
                # try to do it the easy way
                hdf[k] = v
                result = True
            except TypeError:
                result = False
        return result

    def _generic_to_hdf(self, value, hdf, group_name=None):
        """
        Saves also dictionaries and lists to hdf
        Args:
            value: (obj) to object to save
            hdf: the hdf server
            group_name: (str) the group name where to store it
        """
        if isinstance(value, dict):
            # if we deal with a dictionary we have to open a new group anyway
            with hdf.open(group_name) as server:
                # store class metadata
                server['TYPE'] = str(type(value))
                server['FULLNAME'] = fullname(value)
                for k, v in value.items():
                    # try to save it
                    if not isinstance(k, str):
                        # it is possible that the keys are not strings, thus we have to enforce this
                        self.logger.warning('Key "%s" is not a string, it will be converted to %s' % (k, str(k)))
                        k = str(k)
                    # try it the easy way first (either call v.to_hdf or directly save it
                    if self._try_save_key(k, v, server):
                        pass  # everything was successful
                    else:
                        # well pyiron did not manage lets -> more complex object
                        self._generic_to_hdf(v, server, group_name=k)
        elif isinstance(value, (list, tuple)):
            # check if all do have the same type -> then we can make a numpy array out of it
            if len(value) == 0:
                pass  # there is nothing to do, no data to store
            else:
                first_type = type(value[0])
                same = all([type(v) == first_type for v in value])
                # if all items are of the same type and it is simple
                if same and issubclass(first_type, (float, complex, int, np.ndarray)):
                    # that is trivial we do have an array
                    if issubclass(first_type, np.ndarray):
                        # we do not want dtype=object, thus we do make this distinction
                        hdf[group_name] = np.array(value)
                    else:
                        hdf[group_name] = np.array(value, dtype=first_type)
                else:
                    with hdf.open(group_name) as server:
                        # again write the metadata
                        server['TYPE'] = str(type(value))
                        server['FULLNAME'] = fullname(value)
                        for i, v in enumerate(value):
                            index_key = GENERIC_LIST_INDEX_FORMAT.format(index=i)
                            if self._try_save_key(index_key, v, server):
                                pass  # everything was successful
                            else:
                                self._generic_to_hdf(v, server, group_name=index_key)
        else:
            # so this one is the primitive item case
            # lets check if it has a to_hdf method
            try:
                value.to_hdf(hdf, group_name=group_name)
            except AttributeError:
                # Ok there is no to_hdf method however lets try it again
                try:
                    hdf[group_name] = value
                except:
                    # now we have no clue any more, we have to raise this error
                    raise

    def _generic_from_hdf(self, hdf, group_name=None):
        """
        Loads dicts, lists and tuples as well as their subclasses from an hdf file

        Args:
            hdf: the hdf server
            group_name: (str) the group name

        Returns: (obj) the object to return
        """

        # handle special types at first
        # try a simple load
        if 'TYPE' not in hdf[group_name].list_nodes():
            return hdf[group_name]
        elif hdf[group_name]['TYPE'] == str(IODictionary):
            iodict = IODictionary()
            iodict.from_hdf(hdf, group_name)
            return iodict
        elif hdf[group_name]['TYPE'] == str(Atoms):
            struct = Atoms()
            struct.from_hdf(hdf, group_name)
            return struct
        # FULLNAME will only be present if _generic_to_hdf wrote the underlying object
        elif 'FULLNAME' in hdf[group_name].keys():
            with hdf.open(group_name) as server:
                from pydoc import locate
                # convert the class qualifier to a type
                cls_ = locate(server['FULLNAME'])
                # handle a dictionary
                if issubclass(cls_, dict):
                    result = {}
                    # nodes are primitive objects -> that is easy
                    for k in server.list_nodes():
                        # skip the special nodes
                        if k in ('TYPE', 'FULLNAME'):
                            continue
                        result[k] = server[k]

                    for k in server.list_groups():
                        # groups are more difficult, since they're other objects -> give it a try
                        result[k] = self._generic_from_hdf(server, group_name=k)

                    # create the instance -> we have to assume a constructor of type cls_(**kwargs) for that
                    # NOTE: if the default constructor is not available this code will break
                    result = cls_(result)
                    return result
                elif issubclass(cls_, (list, tuple)):
                    result = []
                    # we have to keep track of the indices -> str.__cmp__ != int.__cmp__ we cannot assume an order
                    indices = []

                    for k in server.list_nodes():
                        if k in ('TYPE', 'FULLNAME'):
                            continue
                        # nodes are trivial
                        index = int(k.replace('i_', ''))
                        result.append(server[k])
                        indices.append(index)
                        # TODO: Since Atoms object appear as a node we might have to call it here too

                    for k in server.list_groups():
                        # we do have the recursive call here
                        index = int(k.replace('i_', ''))
                        result.append(self._generic_from_hdf(server, group_name=k))
                        indices.append(index)

                    # sort it, with the keys as indices
                    result = sorted(enumerate(result), key=lambda t: indices[t[0]])
                    # create the instance, and get rid of the instances
                    result = cls_([val for idx, val in result])
                    return result
                else:
                    raise ImportError('Could not locate type(%s)' % server['FULLNAME'])
        else:
            raise TypeError('I do not know how to deserialize type(%s)' % hdf[group_name])

    def to_hdf(self, hdf, group_name=None):
        with hdf.open(group_name) as hdf5_server:
            hdf5_server['TYPE'] = str(type(self))

            for key in list(self.keys()):
                # default value is not to save any property
                try:
                    value = getattr(self, key)
                    try:
                        value.to_hdf(hdf5_server, group_name=key)
                    except AttributeError:
                        self._generic_to_hdf(value, hdf5_server, group_name=key)
                except KeyError:
                    # to_hdf will get called *before* protocols have run, so the pointers in these dictionaries
                    # won't be able to resolve. For now just let it not resolve and don't save it.
                    continue
                except (RuntimeError, OSError):
                    # if a "key" is initialized with a primitive value and the and the graph was already saved
                    # it might happen that the "key" already exists hdf5_server[key] but is of wrong HDF5 type
                    # e.g dataset instead of group. Thus the underlying library will raise an runtime error.
                    # The current workaround now is to try to delete the dataset and rewrite it
                    # TODO: Change to `del hdf5_server[key]` once pyiron.base.generic.hdfio is fixed
                    import posixpath
                    # hdf5_server.h5_path is relative
                    del hdf5_server[posixpath.join(hdf5_server.h5_path, key)]
                    # now we try again
                    try:
                        value.to_hdf(hdf5_server, group_name=key)
                    except AttributeError:
                        self._generic_to_hdf(value, hdf5_server, group_name=key)
                    except Exception:
                        raise

    def from_hdf(self, hdf, group_name):
        with hdf.open(group_name) as hdf5_server:
            for key in hdf5_server.list_nodes():
                if key in ('TYPE', 'FULLNAME'):
                    continue
                # Nodes are leaves, so just save them directly
                # structures will be listed as nodes
                try:
                    setattr(self, key, hdf5_server[key])
                except Exception as e:
                    self.logger.exception('Failed to load "{}"'.format(key), exc_info=e)
                    setattr(self, key, self._generic_from_hdf(hdf5_server, group_name=key))

            for key in hdf5_server.list_groups():
                # Groups are more complex data types with their own depth
                # For now we only treat other IODicts and Atoms (i.e. structures) explicitly.
                setattr(self, key, self._generic_from_hdf(hdf5_server, group_name=key))


class InputDictionary(IODictionary):
    """
    An ``IODictionary`` which is instantiated with a child dictionary to store default values. If a requested item
    can't be found in this dictionary, a default value is sought.
    """

    def __init__(self):
        super(InputDictionary, self).__init__()
        self.default = IODictionary()

    def __getitem__(self, item):
        try:
            return super(InputDictionary, self).__getitem__(item)
        except (KeyError, IndexError):
            return self.default.__getitem__(item)

    def __getattr__(self, item):
        if item == 'default':
            return object.__getattribute__(self, item)
        else:
            return super(InputDictionary, self).__getattr__(item)

    def __setattr__(self, key, value):
        if key == 'default':
            object.__setattr__(self, key, value)
        else:
            super(InputDictionary, self).__setattr__(key, value)

    def keys(self):
        both_keys = set(super(InputDictionary, self).keys()).union(self.default.keys())
        for k in both_keys:
            yield k

    def values(self):
        for k in self.keys():
            yield self[k]

    def items(self):
        # Make sure the dictionary resolve pointers
        for k in self.keys():
            # It resolves the, since we call __getitem__
            yield k, self[k]

    def __iter__(self):
        # Make sure all keys get into the ** unpacking also those from the default dictionary
        return self.keys().__iter__()


class TimelineDict(LoggerMixin, OrderedDict):
    """
        Dictionary which acts as timeline
    """

    def _parse_key(self, k):

        # leftover should contain only a number, try to parse it
        integer_matches = integer_regex.findall(k)
        if len(integer_matches) > 1:
            self.logger.warning('More than one integer was found. I\'ll take the first one')
            integer_matches = [integer_matches[0]]

        try:
            result = int(integer_matches[0])
        except:
            raise KeyError(k)
        else:
            return result

    def keys(self):
        for k in super(TimelineDict, self).keys():
            yield TIMELINE_DICT_KEY_FORMAT.format(time=k)

    def items(self):
        for k, v in zip(self.keys(), super(TimelineDict, self).values()):
            yield k, v

    def _super_getitem(self, k):
        return super(TimelineDict, self).__getitem__(k)

    def _super_keys(self):
        return super(TimelineDict, self).keys()

    @property
    def timeline(self):
        return np.array(list(self._super_keys()))

    @property
    def data(self):
        return np.array(list(self.values()))

    @property
    def array(self):
        return np.array([
            list(self._super_keys()),
            list(self.values())
        ])

    def _check_key_type(self, key):
        if isinstance(key, str):
            time = self._parse_key(key)
        elif isinstance(key, int):
            time = key
        elif isinstance(key, float):
            self.logger.warning('Floating points number are not allowed here. They will be converted to an integer')
            time = int(key)
        else:
            raise TypeError('Only strings of format "%s", integers and floats are allowed as keys'.format(
                TIMELINE_DICT_KEY_FORMAT))
        return time

    def __setitem__(self, key, value):
        super(TimelineDict, self).__setitem__(self._check_key_type(key), value)

    def __getitem__(self, item):
        return super(TimelineDict, self).__getitem__(self._check_key_type(item))



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

from pyiron_contrib.protocol.generic import CompoundVertex, Protocol
from pyiron_contrib.protocol.primitive.one_state import Counter, ExternalHamiltonian, RandomVelocity, Zeros, \
    VerletPositionUpdate, VerletVelocityUpdate
from pyiron_contrib.protocol.primitive.two_state import IsGEq
from pyiron_contrib.protocol.utils import Pointer

"""
Protocol for molecular dynamics.
"""

__author__ = "Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "18 July, 2019"


class MolecularDynamics(CompoundVertex):
    """
    Runs molecular dynamics. This isn't particularly useful as almost every source code/plain job can do this on its
    own, but rather this is intended for testing and teaching. It also serves as a useful starting point for developing
    algorithms with modified dynamics.

    Input attributes:
        ref_job_full_path (str): Path to the pyiron job to use for evaluating forces and energies.
        structure (Atoms): The structure evolve.
        temperature (float): Temperature to run at in K.
        n_steps (int): How many MD steps to run for. (Default is 100.)
        temperature_damping_timescale (float): Langevin thermostat timescale in fs. (Default is None, which runs NVE.)
        time_step (float): MD time step in fs. (Default is 1.)

    Output attributes:
        energy_pot (float): Total potential energy of the system in eV.
        energy_kin (float): Total kinetic energy of the system in eV.
        positions (numpy.ndarray): Atomic positions in angstroms.
        velocities (numpy.ndarray): Atomic velocities in angstroms/fs.
        forces (numpy.ndarray): Atomic forces in eV/angstrom. Note: These are the potential gradient forces; thermostat
            forces (if any) are not saved.
    """

    DefaultWhitelist = {
        'verlet_positions': {
            'output': {
                'positions': 1000,
            },
        },
        'calc_static': {
            'output': {
                'energy_pot': 1,
                'forces': 1000,
            },
        },
        'verlet_velocities': {
            'output': {
                'energy_kin': 1,
                'velocities': 1000,
            },
        },
    }

    def __init__(self, **kwargs):
        super(MolecularDynamics, self).__init__(**kwargs)

        # Protocol defaults
        id_ = self.input.default
        id_.n_steps = 100
        id_.time_step = 1.
        id_.temperature_damping_timescale = None

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.initial_velocity = RandomVelocity()
        g.initial_forces = Zeros()
        g.check_steps = IsGEq()
        g.clock = Counter()
        g.calc_static = ExternalHamiltonian()
        g.verlet_positions = VerletPositionUpdate()
        g.verlet_velocities = VerletVelocityUpdate()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.initial_velocity,
            g.initial_forces,
            g.check_steps, 'false',
            g.verlet_positions,
            g.calc_static,
            g.verlet_velocities,
            g.clock,
            g.check_steps
        )
        g.starting_vertex = g.initial_velocity
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        g.initial_velocity.input.temperature = ip.temperature
        g.initial_velocity.input.masses = ip.structure.get_masses
        g.initial_velocity.input.overheat_fraction = ip.overheat_fraction

        g.initial_forces.input.shape = ip.structure.positions.shape

        g.verlet_positions.input.default.positions = ip.structure.positions
        g.verlet_positions.input.default.velocities = gp.initial_velocity.output.velocities[-1]
        g.verlet_positions.input.default.forces = gp.initial_forces.output.zeros[-1]
        g.verlet_positions.input.positions = gp.verlet_positions.output.positions[-1]
        g.verlet_positions.input.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.input.forces = gp.calc_static.output.forces[-1]
        g.verlet_positions.input.masses = ip.structure.get_masses
        g.verlet_positions.input.time_step = ip.time_step
        g.verlet_positions.input.temperature = ip.temperature
        g.verlet_positions.input.temperature_damping_timescale = ip.temperature_damping_timescale

        g.calc_static.input.ref_job_full_path = ip.ref_job_full_path
        g.calc_static.input.structure = ip.structure
        g.calc_static.input.positions = gp.verlet_positions.output.positions[-1]

        g.verlet_velocities.input.velocities = gp.verlet_positions.output.velocities[-1]
        g.verlet_velocities.input.forces = gp.calc_static.output.forces[-1]
        g.verlet_velocities.input.masses = ip.structure.get_masses
        g.verlet_velocities.input.time_step = ip.time_step
        g.verlet_velocities.input.temperature = ip.temperature
        g.verlet_velocities.input.temperature_damping_timescale = ip.temperature_damping_timescale

        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'energy_pot': ~gp.calc_static.output.energy_pot[-1],
            'energy_kin': ~gp.verlet_velocities.output.energy_kin[-1],
            'positions': ~gp.verlet_positions.output.positions[-1],
            'velocities': ~gp.verlet_velocities.output.velocities[-1],
            'forces': ~gp.calc_static.output.forces[-1]
        }


class ProtocolMD(Protocol, MolecularDynamics):
    pass



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
from pyiron_contrib.protocol.generic import PrimitiveVertex
import numpy as np

"""
Primitive protocols which have two outbound execution edges.
"""


class BoolVertex(PrimitiveVertex):
    """
    This is a class of commands designed to branch the graph based on a binary check. They do not have

    Attributes
    """
    def __init__(self, name=None):
        super(BoolVertex, self).__init__(name=name)
        self.possible_vertex_states = ["true", "false"]
        self.vertex_state = "false"

    def run(self):
        self.command(**self.input.resolve())


class IsGEq(BoolVertex):
    """
    Checks if an input value is greater than or equal to a target threshold. Vertex state switches from 'false' to
    'true' when the target exceeds the threshold.

    Input attributes:
        target (float/int): The value being checked. (Default is numpy infinity.)
        threshold (float/int): What it's being checked against. (Default is zero.)
    """
    def __init__(self, name=None):
        super(IsGEq, self).__init__(name=name)
        self.input.default.target = np.inf
        self.input.default.threshold = 0

    def command(self, target=np.inf, threshold=0):
        if target >= threshold:
            self.vertex_state = "true"
        else:
            self.vertex_state = "false"


class IsLEq(BoolVertex):
    """
    Checks if an input value is less than or equal to a target threshold. Vertex state switches from 'false' to
    'true' when the target exceeds the threshold.

    Input attributes:
        target (float/int): The value being checked. (Default is zero.)
        threshold (float/int): What it's being checked against. (Default is numpy infinity.)
    """
    def __init__(self, name=None):
        super(IsLEq, self).__init__(name=name)
        self.input.default.target = np.inf
        self.input.default.threshold = 0

    def command(self, target, threshold):
        if target <= threshold:
            self.vertex_state = "true"
        else:
            self.vertex_state = "false"


class ModIsZero(BoolVertex):
    """
    Checks if the target value mod some number is zero.

    Input attributes:
        target (int): The value being checked.
        mod (int): The modulo to use.
    """

    def command(self, target, mod):
        if target % mod == 0:
            self.vertex_state = "true"
        else:
            self.vertex_state = "false"


# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
from pyiron_contrib.protocol.generic import PrimitiveVertex
from pyiron_contrib.protocol.utils import IODictionary
import numpy as np
from pyiron.atomistics.job.interactive import GenericInteractive
from pyiron.lammps.lammps import LammpsInteractive
from scipy.constants import physical_constants
from ase.geometry import find_mic, get_distances  # TODO: Wrap things using pyiron functionality
from pyiron import Project
from pyiron_contrib.protocol.utils import ensure_iterable
from os.path import split
from abc import ABC, abstractmethod
from scipy.linalg import toeplitz

"""
Vertices whose present application extends only to finite temperature string-based protocols.

TODO: Naming consistency among vertices, e.g. `all_centroid_positions` vs `centroid_pos_list`.
"""

__author__ = "Raynol Dsouza, Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "20 July, 2019"


class StringDistances(PrimitiveVertex):
    """
    A parent class for vertices which care about the distance from an image to various centroids on the string.
    """
    def __init__(self, name=None):
        super(PrimitiveVertex, self).__init__(name=name)
        self.input.default.eps = 1e-6

    @abstractmethod
    def command(self, *args, **kwargs):
        pass

    @staticmethod
    def check_closest_to_parent(positions, centroid_positions, all_centroid_positions, cell, pbc, eps):
        """
        Checks which centroid the image is closest too, then measures whether or not that closest centroid is sufficiently
        close to the image's parent centroid.

        Args:
            positions (numpy.ndarray): Atomic positions of this image.
            centroid_positions (numpy.ndarray): The positions of the image's centroid.
            all_centroid_positions (list/numpy.ndarray): A list of positions for all centroids in the string.
            cell (numpy.ndarray): The 3x3 cell vectors for pbcs.
            pbc (numpy.ndarray): Three booleans declaring which dimensions have periodic boundary conditions for finding
                the minimum distance convention.
            eps (float): The maximum distance between the closest centroid and the parent centroid to be considered a match
                (i.e. no recentering necessary).

        Returns:
            (bool): Whether the image is closest to its own parent centroid.
        """
        distances = [np.linalg.norm(find_mic(c_pos - positions, cell, pbc)[0]) for c_pos in all_centroid_positions]
        closest_centroid_positions = all_centroid_positions[np.argmin(distances)]
        match_distance = np.linalg.norm(find_mic(closest_centroid_positions - centroid_positions, cell, pbc)[0])
        return match_distance < eps


class StringRecenter(StringDistances):
    """
    If not, the image's positions and forces are reset to match its centroid.

    Input attributes:
        positions (numpy.ndarray): Atomic positions of the image.
        forces (numpy.ndarray): Atomic forces on the image.
        centroid_positions (numpy.ndarray): The positions of the image's centroid.
        centroid_forces (numpy.ndarray): The forces on the image's centroid.
        all_centroid_positions (list/numpy.ndarray): A list of positions for all centroids in the string.
        cell (numpy.ndarray): The 3x3 cell vectors for pbcs.
        pbc (numpy.ndarray): Three booleans declaring which dimensions have periodic boundary conditions for finding
            the minimum distance convention.
        eps (float): The maximum distance between the closest centroid and the parent centroid to be considered a match
            (i.e. no recentering necessary). (Default is 1e-6.)

    Output attributes:
        positions (numpy.ndarray): Either the original positions passed in, or the centroid positions.
        forces (numpy.ndarray): Either the original forces passed in, or the centroid forces.
        recentered (bool): Whether or not the image got recentered.
    """
    def command(self, positions, forces, centroid_positions, centroid_forces, all_centroid_positions, cell, pbc, eps):
        if self.check_closest_to_parent(positions, centroid_positions, all_centroid_positions, cell, pbc, eps):
            return {
                'positions': positions,
                'forces': forces,
                'recentered': False,
            }
        else:
            return {
                'positions': centroid_positions,
                'forces': centroid_forces,
                'recentered': True
            }


class StringReflect(StringDistances):
    """
    If not, the image's positions and forces are reset to match its centroid.

    Input attributes:
        positions (numpy.ndarray): Atomic positions of the image.
        velocities (numpy.ndarray): Atomic velocities of the image.
        previous_positions (numpy.ndarray): Atomic positions of the image from the previous timestep.
        previous_velocities (numpy.ndarray): Atomic velocities of the image from the previous timestep.
        centroid_positions (numpy.ndarray): The positions of the image's centroid.
        centroid_forces (numpy.ndarray): The forces on the image's centroid.
        all_centroid_positions (list/numpy.ndarray): A list of positions for all centroids in the string.
        cell (numpy.ndarray): The 3x3 cell vectors for pbcs.
        pbc (numpy.ndarray): Three booleans declaring which dimensions have periodic boundary conditions for finding
            the minimum distance convention.
        eps (float): The maximum distance between the closest centroid and the parent centroid to be considered a match
            (i.e. no recentering necessary). (Default is 1e-6.)
        forces (numpy.ndarray): Forces corresponding to the input state. (Default is None.)
        previous_forces (numpy.ndarray): Forces corresponding to the previous state. (Default is None.)

    Output attributes:
        positions (numpy.ndarray): Either the original positions passed in, or the previous ones.
        forces (numpy.ndarray): Either the original forces passed in, or the previous ones.
        reflected (bool): Whether or not the image got recentered.
    """
    def __init__(self, name=None):
        super(StringReflect, self).__init__(name=name)
        self.input.default.forces = None
        self.input.default.previous_forces = None

    def command(self, positions, velocities, previous_positions, previous_velocities, centroid_positions,
                all_centroid_positions, cell, pbc, eps, forces, previous_forces):
        if self.check_closest_to_parent(positions, centroid_positions, all_centroid_positions, cell, pbc, eps):
            return {
                'positions': positions,
                'velocities': velocities,
                'forces': forces,
                'reflected': False
            }
        else:
            return {
                'positions': previous_positions,
                'velocities': -previous_velocities,
                'forces': previous_forces,
                'reflected': True
            }


class PositionsRunningAverage(PrimitiveVertex):
    """
    Calculates the running average of input positions at each call.

    Input attributes:
        positions_list (list/numpy.ndarray): The instantaneous position, which will be updated to the running average
        running_average_list (list/numpy.ndarray): List of existing running averages
        cell (numpy.ndarray): The cell of the structure
        pbc (numpy.ndarray): Periodic boundary condition of the structure

    Output attributes:
        running_average_list (list/numpy.ndarray): The updated running average list

    TODO:
        Handle non-static cells, or at least catch them.
        Refactor this so there are list and serial versions equally available
    """

    def __init__(self, name=None):
        super(PositionsRunningAverage, self).__init__(name=name)
        self._divisor = 1

    def command(self, positions_list, running_average_list, relax_endpoints, cell, pbc):
        # On the first step, divide by 2 to average two positions
        self._divisor += 1
        # How much of the current step to mix into the average
        weight = 1. / self._divisor
        running_average_list = np.array(running_average_list)  # Don't modify this input in place

        for i, pos in enumerate(positions_list):
            if (i == 0 or i == len(positions_list) - 1) and not relax_endpoints:
                continue
            else:
                disp = find_mic(pos - running_average_list[i], cell, pbc)[0]
                running_average_list[i] += weight * disp

        return {
            'running_average_list': running_average_list
        }


class CentroidsRunningAverageMix(PrimitiveVertex):
    """
    Mix in the running average of the positions to the centroid, moving the centroid towards that
    running average by a fraction.

    Input attributes:
        mixing_fraction (float): The fraction of the running average to mix into centroid (Default is 0.1)
        centroids_pos_list (list/numpy.ndarray): List of all the centroids along the string
        running_average_list (list/numpy.ndarray): List of running averages
        cell (numpy.ndarray): The cell of the structure
        pbc (numpy.ndarray): Periodic boundary condition of the structure

    Output attributes:
        centroids_pos_list (list/numpy.ndarray): List centroids updated towards the running average

    TODO:
        Re-write Command base class(es) to better handle serial/list/parallel.
    """

    def __init__(self, name=None):
        super(CentroidsRunningAverageMix, self).__init__(name=name)
        self.input.default.mixing_fraction = 0.1

    def command(self, mixing_fraction, centroids_pos_list, running_average_list, cell, pbc):
        centroids_pos_list = np.array(centroids_pos_list)
        for i, cent in enumerate(centroids_pos_list):
            disp = find_mic(running_average_list[i] - cent, cell, pbc)[0]
            update = mixing_fraction * disp
            centroids_pos_list[i] += update

        return {
            'centroids_pos_list': centroids_pos_list
        }


class CentroidsSmoothing(PrimitiveVertex):
    """
    Global smoothing following Vanden-Eijnden and Venturoli (2009). The actual smoothing strength is the product of the
    nominal smoothing strength (`kappa`), the number of images, and the mixing fraction (`dtau`).

    Input Attributes:
        kappa (float): Nominal smoothing strength.
        dtau (float): Mixing fraction (from updating the string towards the moving average of the image positions).
        all_centroid_positions (list/numpy.ndarray): List of all the centroid positions along the string.

    Output Attributes:
        all_centroid_positions (list/numpy.ndarray): List of smoothed centroid positions.
    """
    def command(self, kappa, dtau, all_centroid_positions):
        # Get the smoothing matrix
        n_images = len(all_centroid_positions)
        smoothing_strength = kappa * n_images * dtau
        smoothing_matrix = self._get_smoothing_matrix(n_images, smoothing_strength)
        smoothed_centroid_positions = np.tensordot(smoothing_matrix, np.array(all_centroid_positions), axes=1)

        return {
            'all_centroid_positions': smoothed_centroid_positions
        }

    @staticmethod
    def _get_smoothing_matrix(n_images, smoothing_strength):
        """
        A function that returns the smoothing matrix used in global smoothing.

        Attributes:
            n_images (int): number of images
            smoothing_strength (float): the smoothing penalty

        Returns:
            smoothing_matrix
        """
        toeplitz_rowcol = np.zeros(n_images)
        toeplitz_rowcol[0] = -2
        toeplitz_rowcol[1] = 1
        second_order_deriv = toeplitz(toeplitz_rowcol, toeplitz_rowcol)
        second_order_deriv[0] = np.zeros(n_images)
        second_order_deriv[-1] = np.zeros(n_images)
        smooth_mat_inv = np.eye(n_images) - smoothing_strength * second_order_deriv

        return np.linalg.inv(smooth_mat_inv)


class CentroidsReparameterization(PrimitiveVertex):
    """
    Use linear interpolation to equally space the jobs between the first and last job in 3N dimensional space,
        using a piecewise function

    Input attributes:
        centroids_pos_list (list/numpy.ndarray): List of all the centroids along the string
        cell (numpy.ndarray): The cell of the structure
        pbc (numpy.ndarray): Periodic boundary condition of the structure

    Output attributes:
        centroids_pos_list (list/numpy.ndarray): List of equally spaced centroids
    """

    def __init__(self, name=None):
        super(CentroidsReparameterization, self).__init__(name=name)

    def command(self, centroids_pos_list, cell, pbc):
        # How long is the piecewise parameterized path to begin with?
        lengths = self._find_lengths(centroids_pos_list, cell, pbc)
        length_tot = lengths[-1]
        length_per_frame = length_tot / (len(centroids_pos_list) - 1)

        # Find new positions for the re-parameterized jobs
        new_positions = [centroids_pos_list[0]]
        for n_left, cent in enumerate(centroids_pos_list[1:-1]):
            n = n_left + 1
            length_target = n * length_per_frame

            # Find the last index not in excess of the target length
            try:
                all_not_over = np.argwhere(lengths < length_target)
                highest_not_over = np.amax(all_not_over)
            except ValueError:
                # If all_not_over is empty
                highest_not_over = 0

            # Interpolate from the last position not in excess
            start = centroids_pos_list[highest_not_over]
            end = centroids_pos_list[highest_not_over + 1]
            disp = find_mic(end - start, cell, pbc)[0]
            interp_dir = disp / np.linalg.norm(disp)
            interp_mag = length_target - lengths[highest_not_over]

            new_positions.append(start + interp_mag * interp_dir)
        new_positions.append(centroids_pos_list[-1])

        # Apply the new positions all at once
        centroids_pos_list = new_positions

        return {
            'centroids_pos_list': centroids_pos_list
        }

    @staticmethod
    def _find_lengths(a_list, cell, pbc):
        """
        Finds the cummulative distance from job to job.

        Attribute:
            a_list (list/numpy.ndarray): List of positions whose lengths are to be calculated
            cell (numpy.ndarray): The cell of the structure
            pbc (numpy.ndarray): Periodic boundary condition of the structure

        Returns:
            lengths (list): Lengths of the positions in the list
        """
        length_cummulative = 0
        lengths = [length_cummulative]
        # First length is zero, all other lengths are wrt the first position in the list
        for n_left, term in enumerate(a_list[1:]):
            disp = find_mic(term - a_list[n_left], cell, pbc)[0]
            length_cummulative += np.linalg.norm(disp)
            lengths.append(length_cummulative)
        return lengths


class MilestoningVertex(PrimitiveVertex):
    """
    Checks whether positions are closest to its own centroid. If not, reverses the velocities, and resets the
        positions and forces to the values from the previous step. Logs total reflections, reflections off edges
        and time spent at each edge

    Input attributes:
        positions_list (list/numpy.ndarray): List of current positions
        velocities_list (list/numpy.ndarray): List of current velocities
        forces_list (list/numpy.ndarray): List of current forces
        previous_positions_list (list/numpy.ndarray): List of previous positions
        prev_velocities_list (list/numpy.ndarray): List of previous velocities
        prev_forces_list (list/numpy.ndarray): List of previous forces
        centroids_pos_list (list/numpy.ndarraymile.graph.milestone): List of all the centroids along the string
        thermalization_steps (int): Number of steps after which to start tracking reflections.
        cell (numpy.ndarray): The cell of the structure
        pbc (numpy.ndarray): Periodic boundary condition of the structure

    Output attributes:
        positions_list (list/numpy.ndarray): The modified positions
        velocities_list (list/numpy.ndarray): The modified velocities
        forces_list (list/numpy.ndarray): The modified forces
        reflections_matrix (numpy.ndarray): The reflections matrix
        edge_reflections_matrix (numpy.ndarray): The edge reflections matrix
        edge_time_matrix (numpy.ndarray): The edge time matrix
    """

    def __init__(self, name=None):
        super(MilestoningVertex, self).__init__(name=name)
        # initialize matrices to be tracked
        self.tracker_list = None
        self.reflections_matrix = None
        self.edge_reflections_matrix = None
        self.edge_time_matrix = None

    def command(self, positions_list, velocities_list, forces_list, prev_positions_list, prev_velocities_list,
                prev_forces_list, all_centroid_positions, thermalization_steps, cell, pbc):
        n_images = len(all_centroid_positions)
        n_edges = int(n_images * (n_images - 1) / 2)
        current_step = self.archive.clock

        reflected_positions = []
        reflected_velocities = []
        reflected_forces = []

        if current_step == 0:
            # form the matrices
            self.tracker_list = [[None, None] for _ in np.arange(n_images)]
            self.reflections_matrix = np.zeros((n_images, n_images))
            self.edge_reflections_matrix = np.array([np.zeros((n_edges, n_edges)) for _ in np.arange(n_images)])
            self.edge_time_matrix = np.array([np.zeros(n_edges) for _ in np.arange(n_images)])

        # Find distance between positions and each of the centroids
        for n, (pos, cent) in enumerate(zip(positions_list, all_centroid_positions)):
            closest_centroid_id = self.get_closest_centroid_index(pos, all_centroid_positions, cell, pbc)
            tracker = self.tracker_list[n]

            if closest_centroid_id == n:
                # Return current positions, velocities and forces
                reflected_positions.append(positions_list[n])
                reflected_velocities.append(velocities_list[n])
                reflected_forces.append(forces_list[n])

                if current_step >= thermalization_steps:
                    # Start reflection tracking
                    if tracker[1] is not None:  # If no reflections, increment time
                        self.edge_time_matrix[n][tracker[1]] += 1
                    # End reflection tracking
            else:
                # Update previous positions, velocities and forces, if positions are not closest to parent centroid
                reflected_positions.append(prev_positions_list[n])
                reflected_velocities.append(-prev_velocities_list[n])
                reflected_forces.append(prev_forces_list[n])

                if current_step >= thermalization_steps:
                    # Start reflection tracking
                    self.reflections_matrix[n, closest_centroid_id] += 1  # Save the reflection
                    indices = np.zeros((n_images, n_images))  # images x images
                    indices[n, closest_centroid_id] = 1  # Record the edge
                    ind = np.tril(indices) + np.triu(indices).T  # Convert to triangular matrix
                    # Record the index of the edge (N_j)
                    n_j = int(np.nonzero(ind[np.tril_indices(n_images, k=-1)])[0][0])

                    if tracker[1] is None:
                        tracker[1] = n_j  # Initialize N_j
                    elif tracker[1] == n_j:
                        self.edge_time_matrix[n][n_j] += 1
                    else:
                        tracker[0] = tracker[1]  # If reflecting off a different edge, change N_j to N_i
                        tracker[1] = n_j  # Set new N_j
                        self.edge_reflections_matrix[n][tracker[0], tracker[1]] += 1
                        # End reflection tracking

        return {
            'positions_list': reflected_positions,
            'velocities_list': reflected_velocities,
            'forces_list': reflected_forces,
            'reflections_matrix': self.reflections_matrix,
            'edge_reflections_matrix': self.edge_reflections_matrix,
            'edge_time_matrix': self.edge_time_matrix
        }

    @staticmethod
    def get_closest_centroid_index(positions, all_centroid_positions, cell, pbc):
        distances = [np.linalg.norm(find_mic(c_pos - positions, cell, pbc)[0]) for c_pos in all_centroid_positions]
        return np.argmin(distances)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
from pyiron_contrib.protocol.generic import PrimitiveVertex
from pyiron_contrib.protocol.utils import Pointer
import numpy as np
from pyiron.atomistics.job.interactive import GenericInteractive
from pyiron.lammps.lammps import LammpsInteractive
from scipy.constants import physical_constants
from ase.geometry import find_mic, get_distances  # TODO: Wrap things using pyiron functionality
from pyiron import Project
from pyiron_contrib.protocol.utils import ensure_iterable
from os.path import split
from abc import ABC, abstractmethod
from pyiron_contrib.protocol.math import welford_online

KB = physical_constants['Boltzmann constant in eV/K'][0]
EV_TO_U_ANGSQ_PER_FSSQ = 0.00964853322  # https://www.wolframalpha.com/input/?i=1+eV+in+u+*+%28angstrom%2Ffs%29%5E2
U_ANGSQ_PER_FSSQ_TO_EV = 1. / EV_TO_U_ANGSQ_PER_FSSQ

"""
Primitive vertices which have only one outbound execution edge.

TODO: Rely on pyiron units instead of hard-coding in scipy's Boltzmann constant in eV/K.
"""

__author__ = "Liam Huber, Raynol Dsouza, Dominik Noeger"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "20 July, 2019"


class BuildMixingPairs(PrimitiveVertex):
    """
    Builds an array of mixing parameters [lambda, (1-lambda)].

    Input attributes:
        n_lambdas (int): How many mixing pairs to create.
        lambdas (list/numpy.ndarray): The individual lambda values to use for the first member of each pair.

    Note:
        Exactly one of `n_lambdas` or `lambdas` must be provided.

    Output attributes:
        lambda_pairs (numpy.ndarray): The (`n_lambdas`, 2)-shaped array of mixing pairs.
    """
    def __init__(self, name=None):
        super(BuildMixingPairs, self).__init__(name=name)
        self.input.default.n_lambdas = None
        self.input.default.custom_lambdas = None

    def command(self, n_lambdas, custom_lambdas):
        if custom_lambdas is not None:
            lambdas = np.array(custom_lambdas)
        else:
            lambdas = np.linspace(0, 1, num=n_lambdas)
        return {
            'lambda_pairs': np.array([lambdas, 1 - lambdas]).T
        }


class Counter(PrimitiveVertex):
    """
    Increments by one at each execution.

    Output attributes:
        n_counts (int): How many executions have passed. (Default is 0.)
    """
    def __init__(self, name=None):
        super(Counter, self).__init__(name=name)
        self.output.n_counts = [0]

    def command(self):
        return {
             'n_counts': self.output.n_counts[-1] + 1
         }


class Compute(PrimitiveVertex):

    def __init__(self, name=None):
        super(Compute, self).__init__(name=name)

    def command(self, *args, **kwargs):
        function_ = self.input.function
        return function_(*args, **kwargs)


class DeleteAtom(PrimitiveVertex):
    """
    Given a structure, deletes one of the atoms.

    Input attributes:
        structure (Atoms): The structure to delete an atom of.
        id (int): Which atom to delete.

    Output attributes:
        structure (Atoms): The new, modified structure.
        mask (numpy.ndarray): The integer ids shared by both the old and new structure.
    """
    def command(self, structure, id):
        vacancy_structure = structure.copy()
        vacancy_structure.pop(id)
        mask = np.delete(np.arange(len(structure)).astype(int), id)
        # mask = np.arange(len(structure)).astype(int)
        return {
            'structure': vacancy_structure,
            'mask': mask
        }


class ExternalHamiltonian(PrimitiveVertex):
    """
    Manages calls to an external interpreter (e.g. Lammps, Vasp, Sphinx...) to produce energies, forces, and possibly
    other properties.

    The collected output can be expanded beyond forces and energies (e.g. to magnetic properties or whatever else the
    interpreting code produces) by modifying the `interesting_keys` in the input. The property must have a corresponding
    interactive getter for this property.

    Input attributes:
        ref_job_full_path (string): The full path to the hdf5 file of the job to use as a reference template.
        structure (Atoms): The structure for initializing the external Hamiltonian. Overwrites the reference job
            structure when provided. (Default is None, the reference job needs to have its structure set.)
        interesting_keys (list[str]): String codes for output properties of the underlying job to collect. (Default is
            ['forces', 'energy_pot'].)
        positions (numpy.ndarray): New positions to evaluate. Shape must match the shape of the structure. (Not set by
            default, only necessary if positions are being updated.)
    """

    def __init__(self, name=None):
        super(ExternalHamiltonian, self).__init__(name=name)
        self.input.default.structure = None
        self.input.default.interesting_keys = ['forces', 'energy_pot']
        self.input.default.positions = None

        self._fast_lammps_mode = True  # Set to false only to intentionally be slow for comparison purposes
        self._job = None
        self._job_project_path = None
        self._job_name = None

    def command(self, ref_job_full_path, structure, interesting_keys, positions):
        if self._job_project_path is None:
            self._initialize(ref_job_full_path, structure)
        elif self._job is None:
            self._reload()
        elif not self._job.interactive_is_activated():
            self._job.status.running = True
            self._job.interactive_open()
            self._job.interactive_initialize_interface()

        if isinstance(self._job, LammpsInteractive) and self._fast_lammps_mode:
            # Run Lammps 'efficiently'
            if positions is not None:
                self._job.interactive_positions_setter(positions)

            self._job._interactive_lib_command(self._job._interactive_run_command)
        elif isinstance(self._job, GenericInteractive):
            # DFT codes are slow enough that we can run them the regular way and not care
            # Also we might intentionally run Lammps slowly for comparison purposes
            if positions is not None:
                self._job.structure.positions = positions

            self._job.calc_static()
            self._job.run()
        else:
            raise TypeError('Job of class {} is not compatible.'.format(self._job.__class__))

        return {key: self.get_interactive_value(key) for key in interesting_keys}

    def _initialize(self, ref_job_full_path, structure):
        loc = self.get_graph_location()
        name = loc + '_job'
        project_path, ref_job_path = split(ref_job_full_path)
        pr = Project(path=project_path)
        ref_job = pr.load(ref_job_path)
        job = ref_job.copy_to(
            project=pr,
            new_job_name=name,
            input_only=True,
            new_database_entry=True
        )

        if structure is not None:
            job.structure = structure

        if isinstance(job, GenericInteractive):
            job.interactive_open()

            if isinstance(job, LammpsInteractive) and self._fast_lammps_mode:
                # Note: This might be done by default at some point in LammpsInteractive, and could then be removed here
                job.interactive_flush_frequency = 10**10
                job.interactive_write_frequency = 10**10
                self._disable_lmp_output = True

            job.calc_static()
            job.run(run_again=True)
            # TODO: Running is fine for Lammps, but wasteful for DFT codes! Get the much cheaper interface
            #  initialization working -- right now it throws a (passive) TypeError due to database issues
        else:
            raise TypeError('Job of class {} is not compatible.'.format(ref_job.__class__))
        self._job = job
        self._job_name = name
        self._job_project_path = project_path

    def _reload(self):
        pr = Project(path=self._job_project_path)
        self._job = pr.load(self._job_name)
        self._job.interactive_open()
        self._job.interactive_initialize_interface()
        self._job.calc_static()
        self._job.run(run_again=True)

    def get_interactive_value(self, key):
        if key == 'positions':
            val = np.array(self._job.interactive_positions_getter())
        elif key == 'forces':
            val = np.array(self._job.interactive_forces_getter())
        elif key == 'energy_pot':
            val = self._job.interactive_energy_pot_getter()
        elif key == 'cells':
            val = np.array(self._job.interactive_cells_getter())
        else:
            raise NotImplementedError
        return val

    def finish(self):
        super(ExternalHamiltonian, self).finish()
        if self._job is not None:
            self._job.interactive_close()

    def to_hdf(self, hdf=None, group_name=None):
        super(ExternalHamiltonian, self).to_hdf(hdf=hdf, group_name=group_name)
        hdf[group_name]["fastlammpsmode"] = self._fast_lammps_mode
        hdf[group_name]["jobname"] = self._job_name
        hdf[group_name]["jobprojectpath"] = self._job_project_path

    def from_hdf(self, hdf=None, group_name=None):
        super(ExternalHamiltonian, self).from_hdf(hdf=hdf, group_name=group_name)
        self._fast_lammps_mode = hdf[group_name]["fastlammpsmode"]
        self._job_name = hdf[group_name]["jobname"]
        self._job_project_path = hdf[group_name]["jobprojectpath"]


class GradientDescent(PrimitiveVertex):
    """
    Simple gradient descent update for positions in `flex_output` and structure.

    Input attributes:
        gamma0 (float): Initial step size as a multiple of the force. (Default is 0.1.)
        fix_com (bool): Whether the center of mass motion should be subtracted off of the position update. (Default is
            True)
        use_adagrad (bool): Whether to have the step size decay according to adagrad. (Default is False)
        output_displacements (bool): Whether to return the per-atom displacement vector in the output dictionary.

    TODO:
        Fix adagrad bug when GradientDescent is passed as a Serial vertex
    """

    def __init__(self, name=None):
        super(GradientDescent, self).__init__(name=name)
        self.input.default.gamma0 = 0.1
        self.input.default.fix_com = True
        self.input.default.use_adagrad = False
        self._accumulated_force = 0

    def command(self, positions, forces, gamma0, use_adagrad, fix_com, mask=None, masses=None,
                output_displacements=True):
        positions = np.array(positions)
        forces = np.array(forces)

        if mask is not None:
            masked = True
            mask = np.array(mask)
            # Preserve data
            unmasked_positions = positions.copy()

            # Mask input data
            positions = positions[mask]
            forces = forces[mask]
            masses = np.array(masses)[mask]
        else:
            masked = False

        if use_adagrad:
            self._accumulated_force += np.sqrt(np.sum(forces * forces))
            gamma0 /= self._accumulated_force

        pos_change = gamma0 * np.array(forces)

        if fix_com:
            masses = np.array(masses)[:, np.newaxis]
            total_mass = np.sum(masses)
            com_change = np.sum(pos_change * masses, axis=0) / total_mass
            pos_change -= com_change
        # TODO: fix angular momentum

        new_pos = positions + pos_change

        if masked:
            unmasked_positions[mask] = new_pos
            new_pos = unmasked_positions
            disp = np.zeros(unmasked_positions.shape)
            disp[mask] = pos_change
        else:
            disp = pos_change

        if output_displacements:
            return {
                'positions': new_pos,
                'displacements': disp
            }
        else:
            return {
                'positions': new_pos
            }

    def to_hdf(self, hdf=None, group_name=None):
        super(GradientDescent, self).to_hdf(hdf=hdf, group_name=group_name)
        hdf[group_name]["accumulatedforce"] = self._accumulated_force

    def from_hdf(self, hdf=None, group_name=None):
        super(GradientDescent, self).from_hdf(hdf=hdf, group_name=group_name)
        self._accumulated_force = hdf[group_name]["accumulatedforce"]


class HarmonicHamiltonian(PrimitiveVertex):
    """

    """

    def command(self, positions, home_positions, cell, pbc, spring_constant):
        dr = find_mic(positions - home_positions, cell, pbc)[0]
        force = -spring_constant * dr
        energy = 0.5 * np.sum(spring_constant * dr * dr)
        return {
            'forces': force,
            'energy_pot': energy
        }


class InterpolatePositions(PrimitiveVertex):
    """
    Creates interpolated positions between the positions of an initial and final structure.
    Returns only interpolated positions and not structures.

    Input attributes:
        structure_initial (Atoms): The initial structure
        structure_initial (Atoms): The final structure
        n_images (int): Number of structures to interpolate

    Output attributes:
        interpolated_positions (list/numpy.ndarray): A list of (n_images) positions interpolated between
            the positions of the initial and final structures.
    """

    def __init__(self, name=None):
        super(InterpolatePositions, self).__init__(name=name)

    def command(self, structure_initial, structure_final, n_images):
        pos_i = structure_initial.positions
        pos_f = structure_final.positions
        cell = structure_initial.cell
        pbc = structure_initial.pbc
        displacement = find_mic(pos_f - pos_i, cell, pbc)[0]

        interpolated_positions = []
        for n, mix in enumerate(np.linspace(0, 1, n_images)):
            interpolated_positions += [pos_i + (mix * displacement)]

        return {
            'interpolated_positions': interpolated_positions
        }


class LangevinThermostat(PrimitiveVertex):
    """
    Calculates the necessary forces for a Langevin thermostat based on drag and a random kick.

    Input dictionary:
        velocities (numpy.ndarray): The per-atom velocities in angstroms/fs.
        masses (numpy.ndarray): The per-atom masses in atomic mass units.
        temperature (float): The target temperature. (Default is 0 K, which just applies a drag force.)
        damping_timescale (float): Damping timescale. (Default is 100 fs.)
        time_step (float): MD time step in fs. (Default is 1 fs.)
        fix_com (bool): Whether to ensure the net force is zero. (Default is True.)

    Output dictionary:
        forces (numpy.ndarray): The per-atom forces from the thermostat.

    TODO: Make a version that uses uniform random numbers like Lammps does (for speed)
    """

    def __init__(self, name=None):
        super(LangevinThermostat, self).__init__(name=name)
        id = self.input.default
        id.temperature = 0.
        id.damping_timescale = 100.
        id.time_step = 1.
        id.fix_com = True

    def command(self, velocities, masses,
                temperature, damping_timescale, time_step, fix_com):

        # Ensure that masses are a commensurate shape
        masses = np.array(masses)[:, np.newaxis]
        gamma = masses / damping_timescale

        noise = np.sqrt(2 * (gamma / time_step) * KB * temperature) * np.random.randn(*velocities.shape)
        drag = -gamma * velocities
        thermostat_forces = noise + drag

        if fix_com:  # Apply zero net force
            thermostat_forces -= np.mean(noise, axis=0)

        return {
            'forces': thermostat_forces
        }


class Max(PrimitiveVertex):
    """
    Numpy's amax (except no `out` option).

    Docstrings directly copied from the `numpy docs`_. The `initial` field is renamed `initial_val` to not conflict with
    the input dictionary.

    _`numpy docs`: https://docs.scipy.org/doc/numpy/reference/generated/numpy.amax.html

    Input attributes:
        a (array_like): Input data.
        axis (None/int/tuple of ints): Axis or axes along which to operate. By default, flattened input is used. If
            this is a tuple of ints, the maximum is selected over multiple axes, instead of a single axis or all the
            axes as before.
        keepdims (bool): If this is set to True, the axes which are reduced are left in the result as dimensions with
            size one. With this option, the result will broadcast correctly against the input array. If the default
            value is passed, then *keepdims* will not be passed through to the `amax` method of sub-classes of
            `ndarray`, however any non-default value will be. If the sub-class’ method does not implement *keepdims*
            any exceptions will be raised.

    Output attributes:
        amax (numpy.ndarray/scalar): Maximum of *a*. If axis is None, the result is a scalar value. If axis is given,
            the result is an array of dimension `a.ndim - 1`.

    Note: This misses the new argument ``initial`` in the latest version of Numpy.
    """
    def __init__(self, name=None):
        super(Max, self).__init__(name=name)
        self.input.default.axis = None
        self.input.default.keepdims = False
        self.input.default.initial_val = None

    def command(self, a, axis, keepdims, initial_val):
        return {
            'amax': np.amax(a, axis=axis, keepdims=keepdims)
        }


class NEBForces(PrimitiveVertex):
    """
    Given a list of positions, forces, and energies for each image along some transition, calculates the tangent
    direction along the transition path for each image and returns new forces which have their original value
    perpendicular to this tangent, and 'spring' forces parallel to the tangent. Endpoints forces are set to zero, so
    start with relaxed endpoint structures.

    Note:
        All images must use the same cell and contain the same number of atoms.

    Input attributes:
        positions_list (list/numpy.ndarray): The positions of the images. Each element should have the same shape, and
            the order of the images is relevant!
        energies (list/numpy.ndarray): The potential energies associated with each set of positions. (Not always
            needed.)
        force_list (list/numpy.ndarray): The forces from the underlying energy landscape on which the NEB is being run.
            Must have the same shape as `positions_list`. (Not always needed.)
        cell (numpy.ndarray): The cell matrix the positions live in. All positions must share the same cell.
        pbc (list/numpy.ndarray): Three bools declaring the presence of periodic boundary conditions along the three
            vectors of the cell.
        spring_constant (float): Spring force between atoms in adjacent images. (Default is 1.0 eV/angstrom^2.)
        tangent_style ('plain'/'improved'/'upwinding'): How to calculate the image tangent in 3N-dimensional space.
            (Default is 'upwinding', which requires energies to be set.)
        use_climbing_image (bool): Whether to replace the force with one that climbs along the tangent direction for
            the job with the highest energy. (Default is True, which requires energies and a list of forces to be set.)
        smoothing (float): Strength of the smoothing spring when consecutive images form an angle. (Default is None, do
            not apply such a force.)

    Output attributes:
        forces_list (list): The forces after applying nudged elastic band method.

    TODO:
        Add references to papers (Jonsson and others)
        Implement Sheppard's equations for variable cell shape
    """

    def __init__(self, name=None):
        super(NEBForces, self).__init__(name=name)
        self.input.default.spring_constant = 1.
        self.input.default.tangent_style = 'upwinding'
        self.input.default.use_climbing_image = True
        self.input.default.smoothing = None

    def command(self, positions_list, energies, forces_list, cell, pbc,
                spring_constant, tangent_style, smoothing, use_climbing_image):
        if use_climbing_image:
            climbing_image_index = np.argmax(energies)
        else:
            climbing_image_index = None

        n_images = len(positions_list)
        neb_forces = []
        for i, pos in enumerate(positions_list):
            if i == 0 or i == n_images - 1:
                neb_forces.append(np.zeros(pos.shape))
                continue

            # Otherwise calculate the spring forces
            pos_left = positions_list[i - 1]
            pos_right = positions_list[i + 1]

            # Get displacement to adjacent images
            dr_left = find_mic(pos - pos_left, cell, pbc)[0]
            dr_right = find_mic(pos_right - pos, cell, pbc)[0]

            # Get unit vectors to adjacent images
            tau_left = dr_left / np.linalg.norm(dr_left)
            tau_right = dr_right / np.linalg.norm(dr_right)

            # Calculate the NEB tangent vector
            if tangent_style == 'plain':
                tau = self.normalize(dr_right + dr_left)
            elif tangent_style == 'improved':
                tau = self.normalize(tau_left + tau_right)
            elif tangent_style == 'upwinding':
                en_left = energies[i - 1]
                en = energies[i]
                en_right = energies[i + 1]
                tau = self._get_upwinding_tau(tau_left, tau_right, en_left, en, en_right)
            else:
                # This branch should be inaccessible due to the setter...
                raise KeyError("No such tangent_style: " + str(tangent_style))

            if smoothing is not None:
                force_smoothing = (self.saturating_angle_control(dr_left, dr_right) *
                                   smoothing * (dr_right - dr_left))
            else:
                force_smoothing = 0

            # Decompose the original forces
            input_force = forces_list[i]
            input_force_parallel = np.sum(input_force * tau) * tau
            input_force_perpendicular = input_force - input_force_parallel

            # If climbing image is activated, push the highest energy image up along tau while
            # relaxing otherwise
            if use_climbing_image and i == climbing_image_index:
                neb_forces.append(input_force_perpendicular - input_force_parallel)
            else:
                dr_mag = np.linalg.norm(dr_right) - np.linalg.norm(dr_left)
                force_spring = spring_constant * dr_mag * tau
                neb_forces.append(input_force_perpendicular + force_spring + force_smoothing)

        return {
            'forces_list': neb_forces
        }

    def _get_upwinding_tau(self, tau_left, tau_right, en_left, en, en_right):
        """
        Take direction to the higher-energy image.
        """

        # Find the relative energies
        den_left = en - en_left
        den_right = en - en_right

        # Create booleans to describe the energetic position of the job
        high_left = en_left > en_right
        high_right = en_right > en_left
        extrema = (en_right > en < en_left) or (en_right < en > en_left)

        den_mag_min = min(abs(den_left), abs(den_right))
        den_mag_max = max(abs(den_left), abs(den_right))
        # Calculate the NEB direction
        tau = tau_left * (den_mag_max * high_left + extrema * den_mag_min * high_right) + \
              tau_right * (den_mag_max * high_right + extrema * den_mag_min * high_left)
        return self.normalize(tau)

    @staticmethod
    def normalize(vec):
        l1 = np.linalg.norm(vec)
        if l1 == 0:
            return vec
        return vec / l1

    @staticmethod
    def saturating_angle_control(dr_left, dr_right):
        """
        Return 0 when displacement to left and right neighbours is parallel, and run smoothly to 1
        then saturate at this value when these displacement vectors form an acute angle.
        """
        dr_left_mag = np.linalg.norm(dr_left)
        dr_right_mag = np.linalg.norm(dr_right)
        cos_angle = np.sum(dr_left * dr_right) / (dr_left_mag * dr_right_mag)
        return 0.5 * (1 + np.cos(np.pi * cos_angle))


class Norm(PrimitiveVertex):
    """
    Numpy's linalg norm.

    Docstrings directly copied from the `numpy docs`_.

    _`numpy docs`: https://docs.scipy.org/doc/numpy/reference/generated/numpy.linalg.norm.html

    Input attributes:
        x (array_like): Input array. If axis is None, x must be 1-D or 2-D.
        ord (non-zero int/inf/-inf/'fro'/'nuc'): Order of the norm. inf means numpy’s inf object. (Default is 2).
        axis (int/2-tuple of ints/None): If axis is an integer, it specifies the axis of x along which to compute the
            vector norms. If axis is a 2-tuple, it specifies the axes that hold 2-D matrices, and the matrix norms of
            these matrices are computed. If axis is None then either a vector norm (when x is 1-D) or a matrix norm
            (when x is 2-D) is returned. (Default is None.)
        keepdims (bool): If this is set to True, the axes which are normed over are left in the result as dimensions
            with size one. With this option the result will broadcast correctly against the original x. (Default is
            False)

    Output attributes:
        n (float/numpy.ndarray): Norm of the matrix or vector(s).
    """
    def __init__(self, name=None):
        super(Norm, self).__init__(name=name)
        self.input.default.ord = 2
        self.input.default.axis = None
        self.input.default.keepdims = False

    def command(self, x, ord, axis, keepdims):
        return {
            'n': np.linalg.norm(x, ord=ord, axis=axis, keepdims=keepdims)
        }


class Overwrite(PrimitiveVertex):
    """
    Overwrite particular entries of an array with new values
    """
    def command(self, target, mask, new_values):
        overwritten = np.array(target)
        overwritten[mask] = new_values
        return {
            'overwritten': overwritten
        }


class RandomVelocity(PrimitiveVertex):
    """
    Generates a set of random velocities which (on average) have give the requested temperature.
    Hard-coded for 3D systems.

    Input attributes:
        temperature (float): The temperature of the velocities (in Kelvin).
        masses (numpy.ndarray/list): The masses of the atoms.
        overheat_fraction (float): The fraction by which to overheat the initial velocities. This can be useful for
            more quickly equilibrating a system whose initial structure is its fully relaxed positions -- in which
            case equipartition of energy tells us that the kinetic energy should be initialized to double the
            desired value. (Default is 2.0, assume energy equipartition is a good idea.)

    Output attributes:
        velocities (numpy.ndarray): Per-atom velocities.
        energy_kin (float): Total kinetic energy of all atoms.
    """

    def __init__(self, name=None):
        super(RandomVelocity, self).__init__(name=name)
        self.input.default.overheat_fraction = 2.

    def command(self, temperature, masses, overheat_fraction):

        masses = np.array(masses)[:, np.newaxis]
        vel_scale = np.sqrt(EV_TO_U_ANGSQ_PER_FSSQ * KB * temperature / masses) * np.sqrt(overheat_fraction)
        vel_dir = np.random.randn(len(masses), 3)
        vel = vel_scale * vel_dir
        vel -= np.mean(vel, axis=0)
        energy_kin = 0.5 * np.sum(masses * vel * vel) * U_ANGSQ_PER_FSSQ_TO_EV
        return {
            'velocities': vel,
            'energy_kin': energy_kin
        }


class SphereReflection(PrimitiveVertex):
    """
    Checks whether each atom in a structure is within a cutoff radius of its reference position; if not, reverts the
    positions and velocities of *all atoms* to an earlier state and reverses those earlier velocities. Forces from the
    current and previous time step can be provided optionally.

    Input attributes:
        reference_positions (numpy.ndarray): The reference positions to check the distances from.
        cutoff (float): The cutoff distance from the reference position to trigger reflection.
        positions (numpy.ndarray): The positions to check.
        velocities (numpy.ndarray): The velocities corresponding to the positions at this time.
        previous_positions (numpy.ndarray): The previous positions to revert to.
        previous_velocities (numpy.ndarray): The previous velocities to revert to and reverse.
        pbc (numpy.ndarray/list): Three booleans declaring which dimensions have periodic boundary conditions for
            finding the minimum distance convention.
        cell (numpy.ndarray): The 3x3 cell vectors for pbcs.
        forces (numpy.ndarray): The forces corresponding to the positions at this time. (Default is None.)
        previous_forces (numpy.ndarray): The forces at the `previous_positions` to revert to. (Default is None.)

    Output attributes:
        positions (numpy.ndarray): The (possibly reverted) positions.
        velocities (numpy.ndarray): The (possibly reverted and reversed) velocities.
        forces (numpy.ndarray/None): The (possibly reverted) forces, if force input was provided, else None.
        reflected_mask (bool): Whether or not reflection occurred.
    """

    def __init__(self, name=None):
        super(SphereReflection, self).__init__(name=name)
        self.input.default.forces = None
        self.input.default.previous_forces = None
        # self.input.default.previous_velocities = Pointer(self.input.velocities)

    def command(self, reference_positions, cutoff_distance, positions, velocities, previous_positions,
                previous_velocities, pbc, cell, forces, previous_forces):
        distance = np.linalg.norm(find_mic(reference_positions - positions, cell=cell, pbc=pbc)[0], axis=-1)
        is_at_home = (distance < cutoff_distance)[:, np.newaxis]

        if np.all(is_at_home):
            return {
                'positions': positions,
                'velocities': velocities,
                'forces': forces,
                'reflected': False
            }
        else:
            return {
                'positions': previous_positions,
                'velocities': -previous_velocities,
                'forces': previous_forces,
                'reflected': True
            }


class SphereReflectionPeratom(PrimitiveVertex):
    """
    Checks whether each atom in a structure is within a cutoff radius of its reference position; if not, reverts the
    positions and velocities of *the violating atoms* to an earlier state and reverses those earlier velocities. The
    remaining atoms are unaffected.

    Input attributes:
        reference_positions (numpy.ndarray): The reference positions to check the distances from.
        cutoff (float): The cutoff distance from the reference position to trigger reflection.
        positions (numpy.ndarray): The positions to check.
        velocities (numpy.ndarray): The velocities corresponding to the positions at this time.
        previous_positions (numpy.ndarray): The previous positions to revert to.
        previous_velocities (numpy.ndarray): The previous velocities to revert to and reverse.
        pbc (numpy.ndarray/list): Three booleans declaring which dimensions have periodic boundary conditions for
            finding the minimum distance convention.
        cell (numpy.ndarray): The 3x3 cell vectors for pbcs.

    Output attributes:
        positions (numpy.ndarray): The (possibly reverted) positions.
        velocities (numpy.ndarray): The (possibly reverted and reversed) velocities.
        reflected_mask (numpy.ndarray): A boolean mask that is true for each atom who was reflected.
    """

    def __init__(self, name=None):
        super(SphereReflectionPeratom, self).__init__(name=name)
        # self.input.default.previous_velocities = Pointer(self.input.velocities)

    def command(self, reference_positions, cutoff_distance, positions, velocities, previous_positions,
                previous_velocities, pbc, cell):
        distance = np.linalg.norm(find_mic(reference_positions - positions, cell=cell, pbc=pbc)[0], axis=-1)
        is_at_home = (distance < cutoff_distance)[:, np.newaxis]
        is_away = 1 - is_at_home

        return {
            'positions': is_at_home * positions + is_away * previous_positions,
            'velocities': is_at_home * velocities + is_away * -previous_velocities,
            'reflected': is_away.astype(bool).flatten()
        }


class Slice(PrimitiveVertex):
    def __init__(self, name=None):
        super(Slice, self).__init__(name=name)
        self.input.default.ensure_iterable_mask = False

    def command(self, vector, mask, ensure_iterable_mask):
        if ensure_iterable_mask:
            mask = ensure_iterable(mask)
        return {
            'sliced': vector[mask]
        }


class Transpose(PrimitiveVertex):
    """
    """

    def command(self, matrix):
        transpose = list(map(list, zip(*matrix)))
        return {
            'matrix_transpose': transpose
        }


class VerletParent(PrimitiveVertex, ABC):
    """
    A parent class for holding code which is shared between both position and velocity updates in two-step Velocity
    Verlet.

    TODO: VerletVelocityUpdate should *always* have its velocity input wired to the velocity outupt of
          VerletPositionUpdate. This implies to me that we need some structure *other* than two fully independent
          nodes. It would also be nice to syncronize, e.g. the thermostat and timestep input which is also the same for
          both. However, I haven't figured out how to do that in the confines of the current graph traversal and hdf5
          setup.
    """

    def __init__(self, name=None):
        super(VerletParent, self).__init__(name=name)
        id = self.input.default
        id.time_step = 1.
        id.temperature = None
        id.temperature_damping_timescale = None

    @abstractmethod
    def command(self, *arg, **kwargs):
        pass

    @staticmethod
    def reshape_masses(masses):
        if len(np.array(masses).shape) == 1:
            masses = np.array(masses)[:, np.newaxis]
        return masses

    @staticmethod
    def convert_to_acceleration(forces, masses):
        return forces * EV_TO_U_ANGSQ_PER_FSSQ / masses

    @staticmethod
    def langevin_delta_v(temperature, time_step, masses, damping_timescale, velocities):
        """
        Velocity changes due to the Langevin thermostat.

        Args:
            temperature (float): The target temperature in K.
            time_step (float): The MD time step in fs.
            masses (numpy.ndarray): Per-atom masses in u with a shape (N_atoms, 1).
            damping_timescale (float): The characteristic timescale of the thermostat in fs.
            velocities (numpy.ndarray): Per-atom velocities in angstrom/fs.

        Returns:
            (numpy.ndarray): Per atom accelerations to use for changing velocities.
        """
        drag = -0.5 * time_step * velocities / damping_timescale
        noise = np.sqrt(EV_TO_U_ANGSQ_PER_FSSQ * KB * temperature * time_step / (masses * damping_timescale)) \
            * np.random.randn(*velocities.shape)
        noise -= np.mean(noise, axis=0)
        return drag + noise


class VerletPositionUpdate(VerletParent):
    """
    First half of Velocity Verlet, where positions are updated and velocities are set to their half-step value.

    Input dictionary:
        positions (numpy.ndarray): The per-atom positions in angstroms.
        velocities (numpy.ndarray): The per-atom velocities in angstroms/fs.
        forces (numpy.ndarray): The per-atom forces in eV/angstroms.
        masses (numpy.ndarray): The per-atom masses in atomic mass units.
        time_step (float): MD time step in fs. (Default is 1 fs.)
        temperature (float): The target temperature. (Default is None, no thermostat is used.)
        damping_timescale (float): Damping timescale in fs. (Default is None, no thermostat is used.)

    Output dictionary:
        positions (numpy.ndarray): The new positions on time step in the future.
        velocities (numpy.ndarray): The new velocities *half* a time step in the future.
    """

    def command(self, positions, velocities, forces, masses, time_step, temperature, temperature_damping_timescale):
        masses = self.reshape_masses(masses)
        acceleration = self.convert_to_acceleration(forces, masses)
        vel_half = velocities + 0.5 * acceleration * time_step
        if temperature_damping_timescale is not None:
            vel_half += self.langevin_delta_v(
                temperature,
                time_step,
                masses,
                temperature_damping_timescale,
                velocities
            )
        pos_step = positions + vel_half * time_step

        return {
            'positions': pos_step,
            'velocities': vel_half
        }


class VerletVelocityUpdate(VerletParent):
    """
    Second half of Velocity Verlet, where velocities are updated. Forces should be updated between the position and
    velocity updates

    Input dictionary:
        velocities (numpy.ndarray): The per-atom velocities in angstroms/fs. These should be the half-step velocities
            output by `VerletPositionUpdate`.
        forces (numpy.ndarray): The per-atom forces in eV/angstroms. These should be updated since the last call of
            `VerletPositionUpdate`.
        masses (numpy.ndarray): The per-atom masses in atomic mass units.
        time_step (float): MD time step in fs. (Default is 1 fs.)
        temperature (float): The target temperature. (Default is None, no thermostat is used.)
        damping_timescale (float): Damping timescale in fs. (Default is None, no thermostat is used.)

    Output dictionary:
        velocities (numpy.ndarray): The new velocities *half* a time step in the future.
        energy_kin (float): The total kinetic energy of the system in eV
    """

    def command(self, velocities, forces, masses, time_step, temperature, temperature_damping_timescale):
        masses = self.reshape_masses(masses)
        acceleration = self.convert_to_acceleration(forces, masses)

        vel_step = velocities + 0.5 * acceleration * time_step
        if temperature_damping_timescale is not None:
            vel_step += self.langevin_delta_v(
                temperature,
                time_step,
                masses,
                temperature_damping_timescale,
                velocities
            )
        kinetic_energy = 0.5 * np.sum(masses * vel_step * vel_step) / EV_TO_U_ANGSQ_PER_FSSQ

        return {
            'velocities': vel_step,
            'energy_kin': kinetic_energy
        }


class VoronoiReflection(PrimitiveVertex):
    """
    Checks whether each atom in a structure is closest to its own reference site; if not, reverts the positions and
    velocities to an earlier state and reverses those earlier velocities.

    Input attributes:
        reference_positions (numpy.ndarray): The reference positions to check the distances from.
        positions (numpy.ndarray): The positions to check.
        velocities (numpy.ndarray): The velocities to check.
        previous_positions (numpy.ndarray): The previous positions to revert to.
        previous_velocities (numpy.ndarray): The previous velocities to revert to and reverse.
        pbc (numpy.ndarray/list): Three booleans declaring which dimensions have periodic boundary conditions for
            finding the minimum distance convention.
        cell (numpy.ndarray): The 3x3 cell vectors for pbcs.

    Output attributes:
        positions (numpy.ndarray): The (possibly reverted) positions.
        velocities (numpy.ndarray): The (possibly reverted and reversed) velocities.
        reflected_mask (numpy.ndarray): A boolean mask that is true for each atom who was reflected.
    """

    def __init__(self, name=None):
        super(VoronoiReflection, self).__init__(name=name)
        # self.input.default.previous_velocities = Pointer(self.input.velocities)

    def command(self, reference_positions, positions, velocities, previous_positions, previous_velocities, pbc, cell):
        _, distance_matrix = get_distances(p1=reference_positions, p2=positions, cell=cell, pbc=pbc)
        closest_reference = np.argmin(distance_matrix, axis=0)
        is_at_home = (closest_reference == np.arange(len(positions)))[:, np.newaxis]
        is_away = 1 - is_at_home
        return {
            'positions': is_at_home * positions + is_away * previous_positions,
            'velocities': is_at_home * velocities + is_away * -previous_velocities,
            'reflected': is_away.astype(bool).flatten()
        }


class WeightedSum(PrimitiveVertex):
    """
    Given a list of vectors of with the same shape, calculates a weighted sum of the vectors. By default the weights are
    the inverse of the number of elements and the sum is just the mean.

    Input attributes:
        vectors (list/numpy.ndarray): The vectors to sum. (Masked) vectors must all be of the same length. If the
            the vectors are already in a numpy array, the 0th index should determine the vector.
        weights (list/numpy.ndarray): A 1D list of coefficients (floats) with the same length as ``vectors``. (Default
            is None, which gives the simple mean.)
        masks (list): If not None, a mask must be passed for each vector to extract a sub-vector. The resulting
            collection of vectors and sub-vectors must all have the same length. If the mask for a given vector is
            `None`, all elements are retained. Otherwise the mask must be integer-like or boolean. (Default is None,
            do not mask any of the vectors.)

    Output attributes:
        weighted_sum (numpy.ndarray): The weighted sum, having the same shape as a (masked) input vector.
    """

    def __init__(self, name=None):
        super(WeightedSum, self).__init__(name=name)
        self.input.default.weights = None
        self.input.default.masks = None

    def command(self, vectors, weights, masks):
        # Handle weight defaults
        if weights is None:
            n = len(vectors)
            weights = np.ones() / n
        elif len(weights) != len(vectors):
            raise ValueError('The length of the weights and vectors must be comensurate, but were {} and {}'.format(
                len(weights), len(vectors)))

        # Mask vectors
        if masks is not None:
            for n, mask in enumerate(masks):
                if mask is not None:
                    if isinstance(mask[0], bool) or isinstance(mask[0], np.bool_):
                        mask = np.array(mask, dtype=bool)
                    else:
                        mask = np.array(mask, dtype=int)
                    vectors[n] = vectors[n][mask]

        # Prepare vectors for numpy operation
        vectors = np.array(vectors)

        # Dot vectors
        weighted_sum = np.tensordot(weights, vectors, axes=1)

        # If dot reduces to a single value, recast
        if len(weighted_sum.shape) == 0:
            weighted_sum = float(weighted_sum)

        return {
            'weighted_sum': weighted_sum
        }


class WelfordOnline(PrimitiveVertex):
    def __init__(self, name=None):
        super(WelfordOnline, self).__init__(name=name)
        self.input.default.mean = None
        self.input.default.std = None
        self.input.default.n_samples = None

        op = Pointer(self.output)
        self.input.mean = op.mean[-1]
        self.input.std = op.std[-1]
        self.input.n_samples = op.n_samples[-1]

    def command(self, sample, mean, std, n_samples):
        if mean is None:
            new_mean = sample
            new_std = 0
            n_samples = 0
        else:
            new_mean, new_std = welford_online(sample, mean, std, n_samples)
        return {
            'mean': new_mean,
            'std': new_std,
            'n_samples': n_samples + 1
        }


class Zeros(PrimitiveVertex):
    """
    A numpy vector of zeros that is pointer-compatible.

    Input attributes:
        shape (int/tuple): The shape of the array.

    Output attributes:
        zeros (numpy.ndarray): An array of numpy.float64 zeros.
    """

    def command(self, shape):
        return {
            'zeros': np.zeros(shape)
        }



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
from pyiron_contrib.protocol.generic import Vertex, PrimitiveVertex, CompoundVertex
from pyiron_contrib.protocol.utils import InputDictionary, Pointer
import numpy as np
from abc import abstractmethod
from multiprocessing import Process, Queue
from pyiron.vasp.interactive import VaspInteractive
from pyiron.sphinx.interactive import SphinxInteractive

"""
A command class for running multiple of the same node
"""

__author__ = "Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Jul 23, 2019"


class ListVertex(PrimitiveVertex):
    """
    A base class for making wrappers to run multiple instances of the same protocol. Protocols which inherit from this
    class require the sub-protocol being copied to be passed in at initialization as an argument.

    Attributes to be assigned to the

    Attributes:
        child_type (Command): A class inheriting from ``Command`` with which to create child instances. (Passed as an
            argument at instantiation.)
        children (list): The instances of the child command to execute. These are created automatically (at run-time if
            they don't exist already).
        broadcast (InputDictionary): Input data to be split element-wise across all the child commands. (Entries here
            should always have the same number of entries as there are children, i.e. each child can have its own
            value.)
        direct (InputDictionary): Input data which is to be copied to each child (i.e. all children have the same
            value.)

    Input attributes:
        n_children (int): How many children to create.
    """

    def __init__(self, child_type):
        if not issubclass(child_type, Vertex):
            raise TypeError('ListVertex children must inherit from Protocol.')
        self.children = None  # Ahead of super so the n_history call doesn't trigger the setter and find no children
        super(ListVertex, self).__init__()
        self.child_type = child_type
        self._initialized = False
        self.direct = InputDictionary()
        self.broadcast = InputDictionary()
        self._n_history = None
        self.n_history = 1

    @abstractmethod
    def command(self, n_children):
        pass

    def finish(self):
        for child in self.children:
            child.finish()
        super(ListVertex, self).finish()

    def _initialize(self, n_children):
        children = [self.child_type(name="child_{}".format(n)) for n in range(n_children)]

        # Locate children in graph
        for n, child in enumerate(children):
            child.graph_parent = self
            child.vertex_name = "child_{}".format(n)

        # Link input to input.direct
        for key in list(self.direct.keys()):
            for child in children:
                setattr(child.input, key, getattr(Pointer(self.direct), key))

        # Link input.default to input.direct.default
        for key in list(self.direct.default.keys()):
            for child in children:
                setattr(child.input.default, key, getattr(Pointer(self.direct.default), key))

        # Link input to input.broadcast
        for key in list(self.broadcast.keys()):
            for n, child in enumerate(children):
                setattr(child.input, key, getattr(Pointer(self.broadcast), key)[n])

        # Link input.default to input.broadcast.default
        for key in list(self.broadcast.default.keys()):
            for n, child in enumerate(children):
                setattr(child.input.default, key, getattr(Pointer(self.broadcast.default), key)[n])

        self.children = children
        self._initialized = True

    @property
    def n_history(self):
        return self._n_history

    @n_history.setter
    def n_history(self, n_hist):
        self._n_history = n_hist
        if self.children is not None:
            for child in self.children:
                child.n_history = n_hist

    def _extract_output_data_from_children(self):
        output_keys = list(self.children[0].output.keys())  # Assumes that all the children are the same...
        if len(output_keys) > 0:
            output_data = {}
            for key in output_keys:
                values = []
                for child in self.children:
                    values.append(child.output[key][-1])
                output_data[key] = values
        else:
            output_data = None
        return output_data

    def to_hdf(self, hdf=None, group_name=None):
        super(ListVertex, self).to_hdf(hdf=hdf, group_name=group_name)
        hdf[group_name]['initialized'] = self._initialized
        if self.children is not None:
            with hdf.open(group_name + "/children") as hdf5_server:
                for n, child in enumerate(self.children):
                    child.to_hdf(hdf=hdf5_server, group_name="child" + str(n))

    def from_hdf(self, hdf=None, group_name=None):
        super(ListVertex, self).from_hdf(hdf=hdf, group_name=group_name)
        self._initialized = hdf[group_name]['initialized']
        if self._initialized:
            with hdf.open(group_name + "/children") as hdf5_server:
                children = []
                for n in np.arange(self.input.n_children, dtype=int):
                    child = self.child_type(name="child_{}".format(n))
                    child.from_hdf(hdf=hdf5_server, group_name="child" + str(n))
                    child.graph_parent = self
                    children.append(child)
            self.children = children


class ParallelList(ListVertex):
    """
    A list of commands which are executed in in parallel. The current implementation uses multiprocessing, which creates
    a new python instance and uses pickle to communicate data to it. (That means objects modified in the new processes
    don't also modify the instance we have in the parent python instance!) This all creates some overhead, and so
    parallel is not likely to be worth calling on most

    Attributes:
        queue (multiprocessing.Queue): The queue used for collecting output from the subprocesses. (Instantiation and
            closing are handled by default in `__init__` and `finish`.)
    """
    def __init__(self, child_type):
        super(ParallelList, self).__init__(child_type)
        self.queue = Queue()

    def command(self, n_children):
        """This controls how the commands are run and is about logistics."""
        if self.children is None:
            self._initialize(n_children)

        for child in self.children:
            child.parallel_setup()

        self.queue = Queue()
        # TODO: Instead, check if the queue is open, if not open it (I know how to close, but need to look up opening)

        # Get all the commands running at once
        jobs = []
        for n, child in enumerate(self.children):
            job = Process(target=child.execute_parallel, args=(self.queue, n, child.input.resolve()))
            jobs.append(job)
            job.start()

        # Wait for everything to finish
        for job in jobs:
            job.join()
        # It works fine without this, but I don't fully understand why
        # Since there's no significant cost to add the join, I thought it's safer to be explicit

        # And then grab their output
        # (n.b. the queue might be disordered, but we've made sure we have access to child number)
        for _ in jobs:
            n, child_output = self.queue.get()
            self.children[n].update_and_archive(child_output)

        # Parse output as usual
        output_data = self._extract_output_data_from_children()
        return output_data

    def finish(self):
        super(ParallelList, self).finish()
        self.queue.close()


class SerialList(ListVertex):
    """
    A list of commands which are run in serial.
    """
    def __init__(self, child_type):
        super(SerialList, self).__init__(child_type=child_type)

    def command(self, n_children):
        """This controls how the commands are run and is about logistics."""
        if self.children is None:
            self._initialize(n_children)

        for child in self.children:
            child.execute()

        output_data = self._extract_output_data_from_children()
        return output_data


class AutoList(ParallelList, SerialList):
    """
    Choose between `SerialList` and `ParallelList` depending on the nature of the children.
    """
    def __init__(self, child_type):
        super(AutoList, self).__init__(child_type=child_type)

    def _is_expensive(self):
        try:
            if isinstance(
                self.children[0],
                (VaspInteractive, SphinxInteractive, CompoundVertex)
            ):  # Whitelist types that should be treated in parallel
                return True
            else:  # Everything else is cheap enough to treat in serial
                return False
        except (AttributeError, KeyError, TypeError):
            return False

    def command(self, n_children):
        if self._is_expensive():
            return ParallelList.command(self, n_children)
        else:
            return SerialList.command(self, n_children)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

from pyiron_contrib.protocol.generic import CompoundVertex, Protocol
from pyiron_contrib.protocol.primitive.one_state import Counter, ExternalHamiltonian, WeightedSum, \
    HarmonicHamiltonian, Transpose, RandomVelocity, LangevinThermostat, \
    VerletPositionUpdate, VerletVelocityUpdate, BuildMixingPairs, DeleteAtom, Overwrite, Slice, VoronoiReflection, \
    WelfordOnline, Zeros
from pyiron_contrib.protocol.primitive.two_state import IsGEq, ModIsZero
from pyiron_contrib.protocol.list import SerialList, ParallelList, AutoList
from pyiron_contrib.protocol.utils import Pointer
import numpy as np
import matplotlib.pyplot as plt
from scipy.constants import physical_constants
from abc import ABC, abstractmethod

KB = physical_constants['Boltzmann constant in eV/K'][0]
HBAR = physical_constants['Planck constant over 2 pi in eV s'][0]
ROOT_EV_PER_ANGSTROM_SQUARE_PER_AMU_IN_S = 9.82269385e13
# https://www.wolframalpha.com/input/?i=sqrt((eV)+%2F+((atomic+mass+units)*(angstroms%5E2)))

"""
Protocols for thermodynamic integration using langevin dynamics.
"""

__author__ = "Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "24 July, 2019"


class TILDParent(CompoundVertex, ABC):
    """
    A parent class for thermodynamic integration by langevin dynamics. Mostly just to avoid duplicate code in
    `HarmonicTILD` and `VacancyTILD`.

    Assumes the presence of `build_lambdas`, `average` (for the thermodynamic average of the integrand), `reflect`
    (to keep each atom closest to its own lattice site), and `mix` (to combine the forces from different
    representations).

    # TODO: Make reflection optional; it makes sense for crystals, but maybe not for molecules
    """

    def get_lambdas(self):
        return self.graph.build_lambdas.output.lambda_pairs[-1][:, 0]

    def get_integrand(self):
        integrand = self.graph.average.output
        return integrand.mean[-1], integrand.std[-1] / np.sqrt(integrand.n_samples[-1])

    def plot_integrand(self):
        fig, ax = plt.subplots()
        lambdas = self.get_lambdas()
        thermal_average, standard_error = self.get_integrand()
        ax.plot(lambdas, thermal_average, marker='o')
        ax.fill_between(lambdas, thermal_average - standard_error, thermal_average + standard_error, alpha=0.3)
        ax.set_xlabel("Lambda")
        ax.set_ylabel("dF/dLambda")
        return fig, ax

    def get_free_energy_change(self):
        return np.trapz(x=self.get_lambdas(), y=self.get_integrand()[0])


class HarmonicTILD(TILDParent):
    """
    """
    DefaultWhitelist = {
        'reflect': {
            'output': {
                'positions': 1000,
            },
        },
        'calc_static': {
            'output': {
                'energy_pot': 1,
            },
        },
        'harmonic': {
            'output': {
                'energy_pot': 1,
            },
        },
        'mix': {
            'output': {
                'weighted_sum': 1,
            }
        },
        'verlet_velocities': {
            'output': {
                'energy_kin': 1,
                'velocities': 1000,
            },
        },
    }

    def __init__(self, **kwargs):
        super(HarmonicTILD, self).__init__(**kwargs)

        id_ = self.input.default
        id_.n_steps = 100
        id_.time_step = 1.
        id_.temperature_damping_timescale = None
        id_.overheat_fraction = 2.
        id_.fix_com = True
        id_.use_reflection = True
        # TODO: Need more than input and default, but rather access order, to work without reflection...
        id_.custom_lambdas = None

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.build_lambdas = BuildMixingPairs()
        g.initial_forces = Zeros()
        g.initial_velocity = SerialList(RandomVelocity)
        g.check_steps = IsGEq()
        g.verlet_positions = SerialList(VerletPositionUpdate)
        g.reflect = SerialList(VoronoiReflection)
        g.calc_static = AutoList(ExternalHamiltonian)
        g.harmonic = SerialList(HarmonicHamiltonian)
        g.transpose_forces = Transpose()
        g.mix = SerialList(WeightedSum)
        g.verlet_velocities = SerialList(VerletVelocityUpdate)
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.build_lambdas,
            g.initial_forces,
            g.initial_velocity,
            g.check_steps, 'false',
            g.verlet_positions,
            g.reflect,
            g.calc_static,
            g.harmonic,
            g.transpose_forces,
            g.mix,
            g.verlet_velocities,
            g.clock,
            g.check_steps,
        )
        g.starting_vertex = g.build_lambdas
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        g.build_lambdas.input.n_lambdas = ip.n_lambdas
        g.build_lambdas.input.custom_lambdas = ip.custom_lambdas

        g.initial_forces.input.shape = ip.structure.positions.shape

        g.initial_velocity.input.n_children = ip.n_lambdas
        g.initial_velocity.direct.temperature = ip.temperature
        g.initial_velocity.direct.masses = ip.structure.get_masses
        g.initial_velocity.direct.overheat_fraction = ip.overheat_fraction

        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        g.verlet_positions.input.n_children = ip.n_lambdas
        g.verlet_positions.direct.default.positions = ip.structure.positions
        g.verlet_positions.broadcast.default.velocities = gp.initial_velocity.output.velocities[-1]
        g.verlet_positions.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.verlet_positions.broadcast.positions = gp.reflect.output.positions[-1]
        g.verlet_positions.broadcast.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.broadcast.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_positions.direct.masses = ip.structure.get_masses
        g.verlet_positions.direct.time_step = ip.time_step
        g.verlet_positions.direct.temperature = ip.temperature
        g.verlet_positions.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        g.reflect.input.n_children = ip.n_lambdas
        g.reflect.direct.default.previous_positions = ip.structure.positions
        g.reflect.broadcast.default.previous_velocities = gp.initial_velocity.output.velocities[-1]

        g.reflect.direct.reference_positions = ip.structure.positions
        g.reflect.direct.pbc = ip.structure.pbc
        g.reflect.direct.cell = ip.structure.cell
        g.reflect.broadcast.positions = gp.verlet_positions.output.positions[-1]
        g.reflect.broadcast.velocities = gp.verlet_positions.output.velocities[-1]
        g.reflect.broadcast.previous_positions = gp.reflect.output.positions[-1]
        g.reflect.broadcast.previous_velocities = gp.reflect.output.velocities[-1]

        g.calc_static.input.n_children = ip.n_lambdas
        g.calc_static.direct.ref_job_full_path = ip.ref_job_full_path
        g.calc_static.direct.structure = ip.structure
        g.calc_static.broadcast.positions = gp.reflect.output.positions[-1]

        g.harmonic.input.n_children = ip.n_lambdas
        g.harmonic.direct.spring_constant = ip.spring_constant
        g.harmonic.direct.home_positions = ip.structure.positions
        g.harmonic.broadcast.positions = gp.reflect.output.positions[-1]
        g.harmonic.direct.cell = ip.structure.cell
        g.harmonic.direct.pbc = ip.structure.pbc

        g.transpose_forces.input.matrix = [
            gp.calc_static.output.forces[-1],
            gp.harmonic.output.forces[-1]
        ]

        g.mix.input.n_children = ip.n_lambdas
        g.mix.broadcast.vectors = gp.transpose_forces.output.matrix_transpose[-1]
        g.mix.broadcast.weights = gp.build_lambdas.output.lambda_pairs[-1]

        g.verlet_velocities.input.n_children = ip.n_lambdas
        g.verlet_velocities.broadcast.velocities = gp.reflect.output.velocities[-1]
        g.verlet_velocities.broadcast.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_velocities.direct.masses = ip.structure.get_masses
        g.verlet_velocities.direct.time_step = ip.time_step
        g.verlet_velocities.direct.temperature = ip.temperature
        g.verlet_velocities.direct.temperature_damping_timescale = ip.temperature_damping_timescale
        g.verlet_velocities.direct.time_step = ip.time_step

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'job_energy_pot': ~gp.calc_static.output.energy_pot[-1],
            'harmonic_energy_pot': ~gp.harmonic.output.energy_pot[-1],
            'energy_kin': ~gp.verlet_velocities.output.energy_kin[-1],
            'positions': ~gp.reflect.output.positions[-1],
            'velocities': ~gp.verlet_velocities.output.velocities[-1],
            'forces': ~gp.mix.output.weighted_sum[-1],
        }

    def get_classical_harmonic_free_energy(self, temperatures=None):
        """
        Get the total free energy of a harmonic oscillator with this frequency and these atoms. Temperatures are clipped
        at 1 micro-Kelvin.

        Returns:
            float/np.ndarray: The sum of the free energy of each atom.
        """
        if temperatures is None:
            temperatures = self.input.temperature
        temperatures = np.clip(temperatures, 1e-6, np.inf)
        beta = 1. / (KB * temperatures)

        return -3 * len(self.input.structure) * np.log(np.pi / (self.input.spring_constant * beta)) / (2 * beta)

    def get_quantum_harmonic_free_energy(self, temperatures=None):
        """
        Get the total free energy of a harmonic oscillator with this frequency and these atoms. Temperatures are clipped
        at 1 micro-Kelvin.

        Returns:
            float/np.ndarray: The sum of the free energy of each atom.
        """
        if temperatures is None:
            temperatures = self.input.temperature
        temperatures = np.clip(temperatures, 1e-6, np.inf)
        beta = 1. / (KB * temperatures)
        f = 0
        for m in self.input.structure.get_masses():
            hbar_omega = HBAR * np.sqrt(self.input.spring_constant / m) * ROOT_EV_PER_ANGSTROM_SQUARE_PER_AMU_IN_S
            f += (3. / 2) * hbar_omega + ((3. / beta) * np.log(1 - np.exp(-beta * hbar_omega)))
        return f


class ProtocolHarmonicTILD(Protocol, HarmonicTILD):
    pass


class VacancyTILD(TILDParent):
    """

    """

    def __init__(self, project=None, name=None, job_name=None):
        super(VacancyTILD, self).__init__(project=project, name=name, job_name=job_name)

        self.input.default.n_steps = 100
        self.input.default.vacancy_id = 0
        self._vacancy_structure_init = None
        self.input.temperature_damping_timescale = 10.

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.delete_vacancy = DeleteAtom()
        g.build_lambdas = BuildMixingPairs()
        g.random_velocity = SerialList(RandomVelocity)
        g.initial_forces = Zeros()
        g.slice_structure = Slice()
        g.check_steps = IsGEq()
        g.verlet_positions = SerialList(VerletPositionUpdate)
        g.reflect = SerialList(VoronoiReflection)
        g.calc_full = SerialList(ExternalHamiltonian)
        g.slice_positions = SerialList(Slice)
        g.calc_vac = SerialList(ExternalHamiltonian)
        g.slice_harmonic = SerialList(Slice)
        g.harmonic = SerialList(HarmonicHamiltonian)
        g.write_vac_forces = SerialList(Overwrite)
        g.write_harmonic_forces = SerialList(Overwrite)
        g.transpose_lambda = Transpose()
        g.mix = SerialList(WeightedSum)
        g.verlet_velocities = SerialList(VerletVelocityUpdate)
        g.check_thermalized = IsGEq()
        g.check_sampling_period = ModIsZero()
        g.transpose_energies = Transpose()
        g.addition = SerialList(WeightedSum)
        g.average = SerialList(WelfordOnline)
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.delete_vacancy,
            g.build_lambdas,
            g.random_velocity,
            g.initial_forces,
            g.slice_structure,
            g.check_steps, 'false',
            g.verlet_positions,
            g.reflect,
            g.calc_full,
            g.slice_positions,
            g.calc_vac,
            g.slice_harmonic,
            g.harmonic,
            g.write_vac_forces,
            g.write_harmonic_forces,
            g.transpose_lambda,
            g.mix,
            g.verlet_velocities,
            g.check_thermalized, 'true',
            g.check_sampling_period, 'true',
            g.transpose_energies,
            g.addition,
            g.average,
            g.clock,
            g.check_steps,
        )
        g.make_edge(g.check_thermalized, g.clock, 'false')
        g.make_edge(g.check_sampling_period, g.clock, 'false')
        g.starting_vertex = self.graph.delete_vacancy
        g.restarting_vertex = self.graph.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        g.delete_vacancy.input.structure = ip.structure
        g.delete_vacancy.input.id = ip.vacancy_id
        shared_ids = gp.delete_vacancy.output.mask[-1]

        g.build_lambdas.input.n_lambdas = ip.n_lambdas
        g.build_lambdas.input.custom_lambdas = ip.custom_lambdas
        # n_children = graph_pointer.build_lambdas.output.lambda_pairs[-1].__len__
        # This doesn't yet work because utils can't import MethodWrapperType and use it at line 305 until I have py 3.7

        g.random_velocity.input.n_children = ip.n_lambdas  # n_children
        g.random_velocity.direct.temperature = ip.temperature
        g.random_velocity.direct.masses = ip.structure.get_masses
        g.random_velocity.direct.overheat_fraction = ip.overheat_fraction

        g.initial_forces.input.shape = ip.structure.positions.shape

        g.slice_structure.input.vector = ip.structure.positions
        g.slice_structure.input.mask = ip.vacancy_id
        g.slice_structure.input.ensure_iterable_mask = True  # To keep positions (1,3) instead of (3,)

        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        g.verlet_positions.input.n_children = ip.n_lambdas
        g.verlet_positions.direct.default.positions = ip.structure.positions
        g.verlet_positions.broadcast.default.velocities = gp.random_velocity.output.velocities[-1]
        g.verlet_positions.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.verlet_positions.broadcast.positions = gp.reflect.output.positions[-1]
        g.verlet_positions.broadcast.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.broadcast.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_positions.direct.masses = ip.structure.get_masses
        g.verlet_positions.direct.time_step = ip.time_step
        g.verlet_positions.direct.temperature = ip.temperature
        g.verlet_positions.direct.temperature_damping_timescale = ip.temperature_damping_timescale
        g.verlet_positions.direct.time_step = ip.time_step

        g.reflect.input.n_children = ip.n_lambdas
        g.reflect.direct.default.previous_positions = ip.structure.positions
        g.reflect.broadcast.default.previous_velocities = gp.random_velocity.output.velocities[-1]

        g.reflect.direct.reference_positions = ip.structure.positions
        g.reflect.direct.pbc = ip.structure.pbc
        g.reflect.direct.cell = ip.structure.cell
        g.reflect.broadcast.positions = gp.verlet_positions.output.positions[-1]
        g.reflect.broadcast.velocities = gp.verlet_positions.output.velocities[-1]
        g.reflect.broadcast.previous_positions = gp.reflect.output.positions[-1]
        g.reflect.broadcast.previous_velocities = gp.verlet_velocities.output.velocities[-1]

        g.calc_full.input.n_children = ip.n_lambdas  # n_children
        g.calc_full.direct.ref_job_full_path = ip.ref_job_full_path
        g.calc_full.direct.structure = ip.structure
        g.calc_full.broadcast.positions = gp.reflect.output.positions[-1]

        g.slice_positions.input.n_children = ip.n_lambdas
        g.slice_positions.broadcast.vector = gp.reflect.output.positions[-1]
        g.slice_positions.direct.mask = shared_ids

        g.calc_vac.input.n_children = ip.n_lambdas  # n_children
        g.calc_vac.direct.ref_job_full_path = ip.ref_job_full_path
        g.calc_vac.direct.structure = gp.delete_vacancy.output.structure[-1]
        g.calc_vac.broadcast.positions = gp.slice_positions.output.sliced[-1]

        g.slice_harmonic.input.n_children = ip.n_lambdas
        g.slice_harmonic.broadcast.vector = gp.reflect.output.positions[-1]
        g.slice_harmonic.direct.mask = ip.vacancy_id
        g.slice_harmonic.direct.ensure_iterable_mask = True

        g.harmonic.input.n_children = ip.n_lambdas
        g.harmonic.direct.spring_constant = ip.spring_constant
        g.harmonic.direct.home_positions = gp.slice_structure.output.sliced[-1]
        g.harmonic.broadcast.positions = gp.slice_harmonic.output.sliced[-1]
        g.harmonic.direct.cell = ip.structure.cell
        g.harmonic.direct.pbc = ip.structure.pbc

        g.write_vac_forces.input.n_children = ip.n_lambdas
        g.write_vac_forces.broadcast.target = gp.calc_full.output.forces[-1]
        g.write_vac_forces.direct.mask = shared_ids
        g.write_vac_forces.broadcast.new_values = gp.calc_vac.output.forces[-1]

        g.write_harmonic_forces.input.n_children = ip.n_lambdas
        g.write_harmonic_forces.broadcast.target = gp.write_vac_forces.output.overwritten[-1]
        g.write_harmonic_forces.direct.mask = ip.vacancy_id
        g.write_harmonic_forces.broadcast.new_values = gp.harmonic.output.forces[-1]

        g.transpose_lambda.input.matrix = [
            gp.calc_full.output.forces[-1],
            gp.write_harmonic_forces.output.overwritten[-1]
        ]

        g.mix.input.n_children = ip.n_lambdas
        g.mix.broadcast.vectors = gp.transpose_lambda.output.matrix_transpose[-1]
        g.mix.broadcast.weights = gp.build_lambdas.output.lambda_pairs[-1]

        g.verlet_velocities.input.n_children = ip.n_lambdas
        g.verlet_velocities.broadcast.velocities = gp.reflect.output.velocities[-1]
        g.verlet_velocities.broadcast.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_velocities.direct.masses = ip.structure.get_masses
        g.verlet_velocities.direct.time_step = ip.time_step
        g.verlet_velocities.direct.temperature = ip.temperature
        g.verlet_velocities.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        g.check_thermalized.input.target = gp.clock.output.n_counts[-1]
        g.check_thermalized.input.threshold = ip.thermalization_steps

        g.check_sampling_period.input.target = gp.clock.output.n_counts[-1]
        g.check_sampling_period.input.default.mod = Pointer(self.archive.period)
        g.check_sampling_period.input.mod = ip.sampling_period

        g.transpose_energies.input.matrix = [
            gp.calc_vac.output.energy_pot[-1],
            gp.harmonic.output.energy_pot[-1],
            gp.calc_full.output.energy_pot[-1]
        ]

        g.addition.input.n_children = ip.n_lambdas
        g.addition.broadcast.vectors = gp.transpose_energies.output.matrix_transpose[-1]
        g.addition.direct.weights = [1, 1, -1]

        g.average.input.n_children = ip.n_lambdas
        g.average.broadcast.sample = gp.addition.output.weighted_sum[-1]

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])


class HarmonicTILDParallel(HarmonicTILD):
    DefaultWhitelist = {}

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.build_lambdas = BuildMixingPairs()
        g.run_lambda_points = ParallelList(HarmonicallyCoupled)

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.build_lambdas,
            g.run_lambda_points
        )
        g.starting_vertex = g.build_lambdas
        g.restarting_vertex = g.run_lambda_points

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        g.build_lambdas.input.n_lambdas = ip.n_lambdas
        g.build_lambdas.input.custom_lambdas = ip.custom_lambdas

        g.run_lambda_points.input.n_children = ip.n_lambdas
        g.run_lambda_points.direct.n_steps = ip.n_steps
        g.run_lambda_points.direct.temperature = ip.temperature
        g.run_lambda_points.direct.masses = ip.structure.get_masses
        g.run_lambda_points.direct.overheat_fraction = ip.overheat_fraction
        g.run_lambda_points.direct.threshold = ip.n_steps
        g.run_lambda_points.direct.ref_job_full_path = ip.ref_job_full_path
        g.run_lambda_points.direct.structure = ip.structure
        g.run_lambda_points.direct.spring_constant = ip.spring_constant
        g.run_lambda_points.broadcast.coupling_weights = gp.build_lambdas.output.lambda_pairs[-1]
        g.run_lambda_points.direct.temperature_damping_timescale = ip.temperature_damping_timescale
        g.run_lambda_points.direct.time_step = ip.time_step
        g.run_lambda_points.direct.fix_com = ip.fix_com
        g.run_lambda_points.direct.use_reflection = ip.use_reflection

    def get_output(self):
        o = Pointer(self.graph.run_lambda_points.output)
        return {
            'job_energy_pot': ~o.job_energy_pot[-1],
            'harmonic_energy_pot': ~o.harmonic_energy_pot[-1],
            'energy_kin': ~o.energy_kin[-1],
            'positions': ~o.positions[-1],
            'velocities': ~o.velocities[-1],
            'forces': ~o.forces[-1],
        }


class ProtocolHarmonicTILDParallel(Protocol, HarmonicTILDParallel):
    pass


class HarmonicallyCoupled(CompoundVertex):
    DefaultWhitelist = {
        'reflect': {
            'output': {
                'positions': 1000,
            },
        },
        'calc_static': {
            'output': {
                'energy_pot': 1,
            },
        },
        'harmonic': {
            'output': {
                'energy_pot': 1,
            },
        },
        'mix': {
            'output': {
                'weighted_sum': 1,
            }
        },
        'verlet_velocities': {
            'output': {
                'energy_kin': 1,
                'velocities': 1000,
            },
        },
    }

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.initial_velocity = RandomVelocity()
        g.initial_forces = Zeros()
        g.check_steps = IsGEq()
        g.verlet_positions = VerletPositionUpdate()
        g.reflect = VoronoiReflection()
        g.calc_static = ExternalHamiltonian()
        g.harmonic = HarmonicHamiltonian()
        g.mix = WeightedSum()
        g.verlet_velocities = VerletVelocityUpdate()
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.initial_velocity,
            g.initial_forces,
            g.check_steps, 'false',
            g.verlet_positions,
            g.reflect,
            g.calc_static,
            g.harmonic,
            g.mix,
            g.verlet_velocities,
            g.clock,
            g.check_steps,
        )
        g.starting_vertex = g.initial_velocity
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        g.initial_velocity.input.temperature = ip.temperature
        g.initial_velocity.input.masses = ip.structure.get_masses
        g.initial_velocity.input.overheat_fraction = ip.overheat_fraction

        g.initial_forces.input.shape = ip.structure.positions.shape

        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        g.verlet_positions.input.default.positions = ip.structure.positions
        g.verlet_positions.input.default.velocities = gp.initial_velocity.output.velocities[-1]
        g.verlet_positions.input.default.forces = gp.initial_forces.output.zeros[-1]
        g.verlet_positions.input.positions = gp.reflect.output.positions[-1]
        g.verlet_positions.input.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.input.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_positions.input.masses = ip.structure.get_masses
        g.verlet_positions.input.time_step = ip.time_step
        g.verlet_positions.input.temperature = ip.temperature
        g.verlet_positions.input.temperature_damping_timescale = ip.temperature_damping_timescale

        g.reflect.on = ip.use_reflection
        g.reflect.input.reference_positions = ip.structure.positions
        g.reflect.input.pbc = ip.structure.pbc
        g.reflect.input.cell = ip.structure.cell
        g.reflect.input.positions = gp.verlet_positions.output.positions[-1]
        g.reflect.input.velocities = gp.verlet_positions.output.velocities[-1]
        g.reflect.input.default.previous_positions = ip.structure.positions
        g.reflect.input.default.previous_velocities = gp.initial_velocity.output.velocities[-1]
        g.reflect.input.previous_positions = gp.reflect.output.positions[-1]
        g.reflect.input.previous_velocities = gp.reflect.output.velocities[-1]

        g.calc_static.input.ref_job_full_path = ip.ref_job_full_path
        g.calc_static.input.structure = ip.structure
        g.calc_static.input.default.positions = gp.verlet_positions.output.positions[-1]
        g.calc_static.input.positions = gp.reflect.output.positions[-1]

        g.harmonic.input.spring_constant = ip.spring_constant
        g.harmonic.input.home_positions = ip.structure.positions
        g.harmonic.input.default.positions = gp.verlet_positions.output.positions[-1]
        g.harmonic.input.positions = gp.reflect.output.positions[-1]
        g.harmonic.input.cell = ip.structure.cell
        g.harmonic.input.pbc = ip.structure.pbc

        g.mix.input.vectors = [
            gp.calc_static.output.forces[-1],
            gp.harmonic.output.forces[-1]
        ]
        g.mix.input.weights = ip.coupling_weights

        g.verlet_velocities.input.default.velocities = gp.verlet_positions.output.velocities[-1]
        g.verlet_velocities.input.velocities = gp.reflect.output.velocities[-1]
        g.verlet_velocities.input.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_velocities.input.masses = ip.structure.get_masses
        g.verlet_velocities.input.time_step = ip.time_step

        self.archive.clock = gp.clock.output.n_counts[-1]
        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'job_energy_pot': ~gp.calc_static.output.energy_pot[-1],
            'harmonic_energy_pot': ~gp.harmonic.output.energy_pot[-1],
            'energy_kin': ~gp.verlet_velocities.output.energy_kin[-1],
            'positions': ~gp.reflect.output.positions[-1],
            'velocities': ~gp.verlet_velocities.output.velocities[-1],
            'forces': ~gp.mix.output.weighted_sum[-1],
        }

    def parallel_setup(self):
        self.graph.calc_static._initialize(self.input.ref_job_full_path, self.input.structure)
